from jaaql.email.email_manager_service import EmailAttachment
from jaaql.mvc.base_model import BaseJAAQLModel, VAULT_KEY__jwt_crypt_key
from jaaql.exceptions.http_status_exception import HttpStatusException, HTTPStatus, ERR__already_installed, ERR__already_signed_up
from jaaql.constants import *
from os.path import join
from jaaql.mvc.response import JAAQLResponse
from jaaql.db.db_interface import DBInterface, RET__rows
from jaaql.interpreter.interpret_jaaql import InterpretJAAQL
from jaaql.mvc.queries import *
from jaaql.utilities.utils import get_jaaql_root
from jaaql.db.db_utils import execute_supplied_statement, execute_supplied_statement_singleton, create_interface, jaaql__encrypt
from jaaql.utilities import crypt_utils
from io import BytesIO
from flask import send_file
import threading
from datetime import datetime, timedelta
from jaaql.mvc.base_model import JAAQLPivotInfo
from os.path import dirname
from jwt.utils import base64url_decode
import uuid
import json
import random
import os
from queue import Queue

ERR__refresh_expired = "Token too old to be used for refresh. Please authenticate again"
ERR__incorrect_install_key = "Incorrect install key!"
ERR__invalid_level = "Invalid level!"
ERR__denied_create_user = "Cannot create that user with your current permissions"
ERR__incorrect_credentials = "Incorrect credentials!"
ERR__lacking_permissions = "Only an administrator can perform this action!"
ERR__schema_invalid = "Schema invalid!"
ERR__cant_send_attachments = "Cannot send attachments to other people"
ERR__keep_alive_failed = "Keep alive failed"
ERR__template_not_signup = "One of the supplied templates is not suitable for signup"
ERR__unexpected_parameters = "Signup data not expected"
ERR__unexpected_validation_column = "Unexpected column in the input parameters '%s'"
ERR__data_validation_table_no_primary = "Data validation table has no primary key"
ERR__cant_find_sign_up = "Cannot locate sign up with key. The key is either incorrect, has expired or has not been activated with the emailed code"
ERR__invite_code_expired = "Invite code expired. Please use the link within the email"
ERR__incorrect_invite_code = "Incorrect invite code"
ERR__template_reset_password = "You cannot use this template to reset your password"
ERR__cant_find_reset = "Cannot locate reset with key. The key is either incorrect, has expired or has not been activated with the emailed code"
ERR__reset_code_expired = "Reset code expired. Please use the link within the email"
ERR__incorrect_reset_code = "Incorrect reset code"
ERR__as_attachment_unexpected = "Input 'as_attachment' unexpected as document is returned as a file link"
ERR__document_created_file = "Document is a file, cannot be downloaded in this way"

PG__default_connection_string = "postgresql://postgres:%s@localhost:5432/jaaql"
DIR__scripts = "scripts"

MODIFIER__allow_conflicts = " ON CONFLICT DO NOTHING"

ATTR__version = "version"
ATTR__count = "count"
ATTR__the_user = "the_user"
ATTR__data_lookup_json = "data_lookup_json"
ATTR__activated = "activated"
ATTR__code_attempts = "code_attempts"
ATTR__closed = "closed"
ATTR__created = "created"
ATTR__expiry_code_ms = "code_expiry_ms"
ATTR__used_key_a = "used_key_a"

SQL__err_duplicate_user = "duplicate key value violates unique constraint \"account_unq_email\""

RESEND__invite_max = 2
RESEND__invite_not_registered_max = 3
RESEND__reset_max = 2
RESET__not_started = 0
RESET__started = 1
RESET__completed = 2

CODE__letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
CODE__invite_length = 5
CODE__max_attempts = 3
CODE__alphanumeric = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"
CODE__reset_length = 8

SIGNUP__not_started = 0
SIGNUP__started = 1
SIGNUP__already_registered = 2
SIGNUP__completed = 3


class JAAQLModel(BaseJAAQLModel):

    VERIFICATION_QUEUE = None
    LOG_QUEUE = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        JAAQLModel.LOG_QUEUE = Queue()
        threading.Thread(target=self.log_thread, args=[JAAQLModel.LOG_QUEUE])

    def check_is_internal_admin(self, connection: DBInterface):
        execute_supplied_statement(connection, QUERY__fetch_internal_admin)

    def log(self, user_id: str, occurred: datetime, duration_ms: int, exception: str, contr_input: str, ip: str, status: int, endpoint: str):
        JAAQLModel.LOG_QUEUE.put({
            KEY__user_id: user_id,
            KEY__occurred: occurred.isoformat(),
            KEY__duration_ms: duration_ms,
            KEY__exception: exception,
            KEY__input: contr_input,
            KEY__ip: ip,
            KEY__status: int(status),
            KEY__endpoint: endpoint
        })

    def log_thread(self, log_queue: Queue):
        while True:
            execute_supplied_statement(self.jaaql_lookup_connection, QUERY__log_ins, log_queue.get())

    def drop_email_account(self, connection: DBInterface, name: str):
        execute_supplied_statement(connection, QUERY__drop_email_account, {KEY__name: name})
        self.email_manager.reload_service()

    def redeploy(self, connection: DBInterface):
        self.check_is_internal_admin(connection)

        if self.is_container:
            f = open("/JAAQL-middleware-python/redeploy", "w")
        else:
            f = open(join(dirname(get_jaaql_root()), "redeploy"), "w")
        f.write("Will be detected and redeployment will now happen")
        f.close()

        print("Redeploying JAAQL")

        threading.Thread(target=self.exit_jaaql, daemon=True).start()

    def add_account_password(self, user_id: str, password: str):
        crypt_utils.validate_password(password)
        salt = self.get_repeatable_salt(user_id)
        password = crypt_utils.hash_password(password, salt)
        password_id = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__add_password,
                                                           {KEY__account: user_id, KEY__password_hash: password},
                                                           encryption_key=self.get_db_crypt_key(), encryption_salts={KEY__password_hash: salt},
                                                           encrypt_parameters=[KEY__password_hash], as_objects=True)[KEY__id]
        execute_supplied_statement(self.jaaql_lookup_connection, QUERY__add_password_cache, {KEY__account: user_id, KEY__password: password_id})

    def add_my_account_password(self, user_id: str, username: str, ip_address: str, old_password: str, password: str):
        res = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_recent_passwords_with_account, {KEY__user_id: user_id},
                                                   decrypt_columns=[KEY__password_hash], encryption_key=self.get_db_crypt_key())
        if not crypt_utils.verify_password_hash(res[KEY__password_hash], old_password, salt=self.get_repeatable_salt(user_id)):
            raise HttpStatusException(ERR__incorrect_credentials)
        self.add_account_password(password, user_id)
        return self.get_auth_token(password=password, ip_address=ip_address, username=username)

    def fetch_user_from_username(self, username: str, singleton_message: str = None, singleton_code: int = None):
        return execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_user_from_username, {KEY__username: username},
                                                    encryption_key=self.get_db_crypt_key(), encrypt_parameters=[KEY__username],
                                                    encryption_salts={KEY__username: self.get_repeatable_salt()}, as_objects=True,
                                                    singleton_code=singleton_code, singleton_message=singleton_message)[KEY__user_id]

    def fetch_user_record_from_username(self, username: str):
        return execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_user_record_from_username, {KEY__username: username},
                                                    encryption_key=self.get_db_crypt_key(), encrypt_parameters=[KEY__username],
                                                    encryption_salts={KEY__username: self.get_repeatable_salt()}, as_objects=True)

    def select_from_data_validation_table(self, connection: DBInterface, val_table: str, pkey_vals: dict):
        val_table_esc = '"%s"' % val_table
        pkeys_where = " AND ".join(['"' + col + '" = :' + col for col in pkey_vals.keys()])
        sel_query = "SELECT * FROM " + val_table_esc + " WHERE " + pkeys_where
        sanitized_params = execute_supplied_statement_singleton(connection, sel_query, pkey_vals, as_objects=True)

        return {key: value for key, value in sanitized_params.items() if key not in pkey_vals}

    def fetch_sanitized_email_params(self, connection: DBInterface, template: dict, params: dict):
        val_table = template[KEY__data_validation_table]
        val_cols = execute_supplied_statement(connection, QUERY__fetch_table_columns,
                                              {KEY__table_name: val_table}, as_objects=True)
        primary_cols = [val_col[KEY__column_name] for val_col in val_cols if val_col[KEY__is_primary]]
        val_cols = [val_col[KEY__column_name] for val_col in val_cols]

        for col in params.keys():
            if col not in val_cols:
                raise HttpStatusException(ERR__unexpected_validation_column % col)

        cols = ", ".join(['"' + col + '"' for col in val_cols if col in params.keys()])
        ins = ", ".join([':' + col for col in val_cols if col in params.keys()])
        pkeys = ", ".join(['"' + col + '"' for col in primary_cols])
        if len(pkeys) == 0:
            raise HttpStatusException(ERR__data_validation_table_no_primary)
        val_table_esc = '"%s"' % val_table
        ins_query = "INSERT INTO " + val_table_esc + " (" + cols + ") VALUES (" + ins + ") RETURNING " + pkeys
        pkey_vals = execute_supplied_statement_singleton(connection, ins_query, params, as_objects=True)
        pkey_vals = {key: str(val) if isinstance(val, uuid.UUID) else val for key, val in pkey_vals.items()}
        select_table = template[KEY__data_validation_view] if template[KEY__data_validation_view] is not None else val_table
        return self.select_from_data_validation_table(connection, select_table, pkey_vals), pkey_vals

    def user_invite(self, user_id: str, application: str = None, configuration: str = None, email_template: str = None,
                    template_lookup: dict = None):
        if template_lookup is not None:
            template_lookup = json.dumps(template_lookup)

        params = {
            ATTR__data_lookup_json: template_lookup,
            ATTR__the_user: user_id,
            KEY__application: application,
            KEY__configuration: configuration,
            KEY__email_template: email_template,
            KEY__invite_code: "".join([CODE__letters[random.randint(0, len(CODE__letters) - 1)] for _ in range(CODE__invite_length)])
        }
        return execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__sign_up_insert, params, as_objects=True)

    def fetch_interface(self, application: str, configuration: str, schema: str):
        """
        TODO this won't work. Used for emails, needs to be the database admin user for a database
        :param application:
        :param configuration:
        :param schema:
        :return:
        """
        schemas = self.fetch_pivoted_application_schemas()
        db_name = schemas[application][configuration][schema][KEY__database]
        return self.create_interface_for_db(ROLE__jaaql, db_name)

    def request_signup(self, inputs: dict):
        template = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_email_template_with_app_url, {
            KEY__name: inputs[KEY__email_template],
            KEY__configuration: inputs[KEY__configuration],
            KEY__application: inputs[KEY__application]
        }, as_objects=True)
        template_already_exists = execute_supplied_statement_singleton(
            self.jaaql_lookup_connection, QUERY__fetch_email_template_with_app_url, {
                KEY__name: inputs[KEY__already_signed_up_email_template],
                KEY__configuration: inputs[KEY__configuration],
                KEY__application: inputs[KEY__application]
            }, as_objects=True
        )
        artifact_url = template.pop(KEY__artifact_base_uri)
        app_url = template.pop(KEY__application_url)

        if not template[KEY__allow_signup] or not template_already_exists[KEY__allow_confirm_signup_attempt]:
            raise HttpStatusException(ERR__template_not_signup)

        user_existed = False
        try:
            user_id = self.create_account(self.jaaql_lookup_connection, {KEY__username: inputs[KEY__email]}, True)
        except HttpStatusException as hs:
            if not hs.message.startswith(SQL__err_duplicate_user):
                raise hs  # Unrelated exception, raise it

            res = self.fetch_user_recent_password(inputs[KEY__email])
            if len(res) == 0:
                user_id = self.fetch_user_from_username(inputs[KEY__email])
            else:
                user_id = res[0][KEY__user_id]
                user_existed = True

        attempts = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__sign_up_count, {KEY__user_id: user_id},
                                                        as_objects=True)[ATTR__count]
        if user_existed and attempts >= RESEND__invite_max or attempts >= RESEND__invite_not_registered_max:
            raise HttpStatusException(ERR__too_many_signup_attempts, HTTPStatus.TOO_MANY_REQUESTS)

        params = inputs[KEY__parameters]
        if params is not None and template[KEY__data_validation_table] is None:
            raise HttpStatusException(ERR__unexpected_parameters)
        if params is None and template[KEY__data_validation_table] is not None:
            params = {}

        template = template_already_exists if user_existed else template

        sanitized_params = {}
        pkey_vals = None
        if template[KEY__data_validation_table] is not None:
            if EMAIL_PARAM__signup_key in params:
                raise HttpStatusException(ERR__unexpected_validation_column % EMAIL_PARAM__signup_key)
            if EMAIL_PARAM__invite_code in params:
                raise HttpStatusException(ERR__unexpected_validation_column % EMAIL_PARAM__invite_code)
            if EMAIL_PARAM__app_url in params:
                raise HttpStatusException(ERR__unexpected_validation_column % EMAIL_PARAM__app_url)

            interface = self.fetch_interface(template[KEY__application], inputs[KEY__configuration], template[KEY__schema])
            sanitized_params, pkey_vals = self.fetch_sanitized_email_params(interface, template, params)

        invite_keys = self.user_invite(user_id, inputs[KEY__application], inputs[KEY__configuration], template[KEY__name], pkey_vals)
        optional_params = {EMAIL_PARAM__signup_key: invite_keys[KEY__invite_key], EMAIL_PARAM__invite_code: invite_keys[KEY__invite_code],
                           EMAIL_PARAM__app_url: app_url}
        optional_params = {**optional_params, **sanitized_params}

        self.email_manager.construct_and_send_email(artifact_url, template, user_id, inputs[KEY__email], None, {}, optional_params)

        return {KEY__invite_key: invite_keys[KEY__invite_poll_key]}

    def signup_status(self, inputs: dict):
        select_param = {KEY__invite_or_poll_key: inputs[KEY__invite_or_poll_key]}
        resp = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__sign_up_poll, parameters=select_param, as_objects=True,
                                                    decrypt_columns=[KEY__email], encryption_key=self.get_db_crypt_key(),
                                                    singleton_message=ERR__cant_find_sign_up)

        res = self.fetch_user_recent_password(resp[KEY__email])
        existed = len(res) != 0

        is_invite_key = str(resp[KEY__invite_key]) == inputs[KEY__invite_or_poll_key]

        if not is_invite_key and not resp[ATTR__activated] and resp[ATTR__code_attempts] >= CODE__max_attempts:
            raise HttpStatusException(ERR__too_many_code_attempts, HTTPStatus.LOCKED)

        status = SIGNUP__not_started

        invite_code_match = inputs.get(KEY__invite_code) == resp[KEY__invite_code]
        timezone = resp[ATTR__created].tzinfo
        invite_code_expired = resp[ATTR__created] + timedelta(milliseconds=resp[ATTR__expiry_code_ms]) < datetime.now(timezone)

        if is_invite_key or resp[ATTR__activated] or (invite_code_match and not invite_code_expired):
            if invite_code_match and not resp[ATTR__activated]:
                execute_supplied_statement(self.jaaql_lookup_connection, QUERY__sign_up_upd, {KEY__invite_key: resp[KEY__invite_key]})
            elif not resp[ATTR__used_key_a]:
                execute_supplied_statement(self.jaaql_lookup_connection, QUERY__sign_up_upd_used, {KEY__invite_key: resp[KEY__invite_key]})

            if resp[ATTR__closed]:
                status = SIGNUP__completed
            elif existed:
                status = SIGNUP__already_registered
            elif resp[ATTR__activated] or resp[ATTR__used_key_a]:
                status = SIGNUP__started

        elif invite_code_match:
            raise HttpStatusException(ERR__invite_code_expired)
        else:
            execute_supplied_statement(self.jaaql_lookup_connection, QUERY__sign_up_increment_attempts, {KEY__invite_key: resp[KEY__invite_key]})
            raise HttpStatusException(ERR__incorrect_invite_code)

        return {KEY__invite_key_status: status}

    def fetch_user_recent_password(self, username: str):
        res = execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fetch_recent_passwords, {
            KEY__enc_username: username
        }, encryption_salts={KEY__enc_username: self.get_repeatable_salt()}, decrypt_columns=[KEY__password_hash, KEY__username],
                                         encrypt_parameters=[KEY__enc_username], encryption_key=self.get_db_crypt_key(), as_objects=True)
        return res

    def sign_up_user_with_token(self, token: str, password: str):
        resp = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__sign_up_fetch, parameters={KEY__invite_or_poll_key: token},
                                                    decrypt_columns=[KEY__email], encryption_key=self.get_db_crypt_key(),
                                                    as_objects=True, singleton_message=ERR__cant_find_sign_up)

        username = resp[KEY__email]

        res = self.fetch_user_recent_password(username)

        if len(res) != 0:
            raise HttpStatusException(ERR__already_signed_up, response_code=HTTPStatus.CONFLICT)

        self.add_account_password(self.fetch_user_from_username(username), password)

        return {KEY__email: username}

    def fetch_signup(self, token: str):
        resp = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__sign_up_fetch, parameters={KEY__invite_or_poll_key: token},
                                                    decrypt_columns=[KEY__email], encryption_key=self.get_db_crypt_key(),
                                                    as_objects=True, singleton_message=ERR__cant_find_sign_up)

        res = {}
        if resp[KEY__email_template] and resp[ATTR__data_lookup_json]:
            template = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_email_template_with_app_url,
                                                            {KEY__name: resp[KEY__email_template],
                                                             KEY__configuration: resp[KEY__configuration], KEY__application: resp[KEY__application]},
                                                            as_objects=True)
            interface = self.fetch_interface(template[KEY__application], resp[KEY__configuration], template[KEY__schema])
            res[KEY__parameters] = self.select_from_data_validation_table(interface, template[KEY__data_validation_table],
                                                                          json.loads(resp[ATTR__data_lookup_json]))

        return res

    def delete_from_data_validation_table(self, connection: DBInterface, val_table: str, pkey_vals: dict):
        val_table_esc = '"%s"' % val_table
        pkeys_where = " AND ".join(['"' + col + '" = :' + col for col in pkey_vals.keys()])
        pkeys_returning = ",".join(['"' + col + '"' for col in pkey_vals.keys()])
        del_query = "DELETE FROM " + val_table_esc + " WHERE " + pkeys_where + " RETURNING " + pkeys_returning
        execute_supplied_statement_singleton(connection, del_query, pkey_vals)

    def finish_signup(self, token: str):
        resp = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__sign_up_fetch, parameters={KEY__invite_or_poll_key: token},
                                                    decrypt_columns=[KEY__email], encryption_key=self.get_db_crypt_key(),
                                                    as_objects=True, singleton_message=ERR__cant_find_sign_up)

        if resp[KEY__email_template] and resp[ATTR__data_lookup_json]:
            template = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_email_template_with_app_url,
                                                            {KEY__name: resp[KEY__email_template],
                                                             KEY__configuration: resp[KEY__configuration]},
                                                            as_objects=True)[KEY__data_validation_table]
            interface = self.fetch_interface(template[KEY__application], resp[KEY__configuration], template[KEY__schema])
            self.delete_from_data_validation_table(interface, template, json.loads(resp[ATTR__data_lookup_json]))

        execute_supplied_statement(self.jaaql_lookup_connection, QUERY__sign_up_close, {KEY__invite_key: resp[KEY__invite_key]})

    def send_reset_password_email(self, inputs: dict):
        template = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_email_template_with_app_url, {
            KEY__name: inputs[KEY__email_template],
            KEY__configuration: inputs[KEY__configuration],
            KEY__application: inputs[KEY__application]
        }, as_objects=True)
        app_url = template.pop(KEY__application_url)
        artifact_url = template.pop(KEY__artifact_base_uri)

        if not template[KEY__allow_reset_password]:
            raise HttpStatusException(ERR__template_reset_password)

        user_exists = True
        user = None
        try:
            user = self.fetch_user_record_from_username(inputs[KEY__email])
        except HttpStatusException:
            user_exists = False

        if user_exists and user[KEY__application] is not None:
            raise HttpStatusException(ERR__user_public, HTTPStatus.UNAUTHORIZED)

        if user_exists:
            attempts = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__reset_count, {ATTR__the_user: user[KEY__user_id]},
                                                            as_objects=True)[ATTR__count]
        else:
            attempts = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fake_reset_count,
                                                            {KEY__email: inputs[KEY__email]},
                                                            encryption_key=self.get_db_crypt_key(),
                                                            encryption_salts={KEY__email: self.get_repeatable_salt()},
                                                            encrypt_parameters=[KEY__email], as_objects=True)[ATTR__count]
        if attempts >= RESEND__reset_max:
            raise HttpStatusException(ERR__too_many_reset_requests, HTTPStatus.TOO_MANY_REQUESTS)

        if user_exists:
            params = inputs[KEY__parameters]
            if params is not None and template[KEY__data_validation_table] is None:
                raise HttpStatusException(ERR__unexpected_parameters)
            if params is None and template[KEY__data_validation_table] is not None:
                params = {}

            sanitized_params = {}
            pkey_vals = None

            if template[KEY__data_validation_table] is not None:
                if EMAIL_PARAM__reset_key in params:
                    raise HttpStatusException(ERR__unexpected_validation_column % EMAIL_PARAM__reset_key)
                if EMAIL_PARAM__reset_code in params:
                    raise HttpStatusException(ERR__unexpected_validation_column % EMAIL_PARAM__reset_code)
                if EMAIL_PARAM__app_url in params:
                    raise HttpStatusException(ERR__unexpected_validation_column % EMAIL_PARAM__app_url)

                interface = self.fetch_interface(template[KEY__application], inputs[KEY__configuration], template[KEY__schema])
                sanitized_params, pkey_vals = self.fetch_sanitized_email_params(interface, template, params)

            template_lookup = None
            if pkey_vals is not None:
                template_lookup = json.dumps(pkey_vals)
            params = {
                ATTR__data_lookup_json: template_lookup,
                ATTR__the_user: user[KEY__user_id],
                KEY__email_template: template[KEY__name],
                KEY__configuration: inputs[KEY__configuration],
                KEY__application: inputs[KEY__application],
                KEY__reset_code: "".join([CODE__alphanumeric[random.randint(0, len(CODE__alphanumeric) - 1)] for _ in range(CODE__reset_length)])
            }

            reset_keys = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__reset_insert, params, as_objects=True)
            optional_params = {EMAIL_PARAM__reset_key: reset_keys[KEY__reset_key], EMAIL_PARAM__reset_code: reset_keys[KEY__reset_code],
                               EMAIL_PARAM__app_url: app_url}
            optional_params = {**optional_params, **sanitized_params}

            self.email_manager.construct_and_send_email(artifact_url, template, user[KEY__user_id], inputs[KEY__email], None, {}, optional_params)
        else:
            params = {
                KEY__email: inputs[KEY__email],
            }
            reset_keys = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fake_reset_insert, params,
                                                              encryption_key=self.get_db_crypt_key(),
                                                              encryption_salts={KEY__email: self.get_repeatable_salt()},
                                                              encrypt_parameters=[KEY__email], as_objects=True)

        return {KEY__reset_key: reset_keys[KEY__reset_poll_key]}

    def reset_password_status(self, inputs: dict):
        select_param = {KEY__reset_or_poll_key: inputs[KEY__reset_or_poll_key]}
        was_fake = False

        try:
            resp = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__reset_poll, parameters=select_param, as_objects=True,
                                                        decrypt_columns=[KEY__email], encryption_key=self.get_db_crypt_key(),
                                                        singleton_message=ERR__cant_find_reset)
        except HttpStatusException as se:
            if se.response_code == HTTPStatus.UNPROCESSABLE_ENTITY:
                resp = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fake_reset_poll, parameters=select_param,
                                                            decrypt_columns=[KEY__email], encryption_key=self.get_db_crypt_key(),
                                                            as_objects=True, singleton_message=ERR__cant_find_reset)
                was_fake = True
            else:
                raise se

        is_reset_key = str(resp[KEY__reset_key]) == inputs[KEY__reset_or_poll_key]

        if not is_reset_key and not resp[ATTR__activated] and resp[ATTR__code_attempts] >= CODE__max_attempts:
            raise HttpStatusException(ERR__too_many_code_attempts, HTTPStatus.LOCKED)

        status = RESET__not_started

        reset_code_match = inputs.get(KEY__reset_code) == resp[KEY__reset_code]
        timezone = resp[ATTR__created].tzinfo
        reset_code_expired = resp[ATTR__created] + timedelta(milliseconds=resp[ATTR__expiry_code_ms]) < datetime.now(timezone)

        if is_reset_key or resp[ATTR__activated] or (reset_code_match and not reset_code_expired):
            if reset_code_match and not resp[ATTR__activated]:
                execute_supplied_statement(self.jaaql_lookup_connection, QUERY__reset_upd, {KEY__reset_key: resp[KEY__reset_key]})
            elif not resp[ATTR__used_key_a]:
                execute_supplied_statement(self.jaaql_lookup_connection, QUERY__reset_upd_used, {KEY__reset_key: resp[KEY__reset_key]})

            if resp[ATTR__closed]:
                status = RESET__completed
            elif resp[ATTR__activated] or resp[ATTR__used_key_a]:
                status = RESET__started

        elif reset_code_match:
            raise HttpStatusException(ERR__reset_code_expired)
        else:
            if was_fake:
                execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fake_reset_increment_attempts,
                                           {KEY__reset_key: inputs[KEY__reset_or_poll_key]})
            else:
                execute_supplied_statement(self.jaaql_lookup_connection, QUERY__reset_increment_attempts, {KEY__reset_key: resp[KEY__reset_key]})
            raise HttpStatusException(ERR__incorrect_reset_code)

        return {KEY__reset_key_status: status}

    def fetch_document(self, inputs: dict, response: JAAQLResponse):
        res = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_rendered_document,
                                                   {KEY__document_id: inputs[KEY__document_id]}, singleton_message=ERR__document_id_not_found,
                                                   as_objects=True)

        if not res[KEY__completed]:
            raise HttpStatusException(ERR__document_still_rendering, HTTP_STATUS__too_early)

        if res[KEY__create_file]:
            if inputs[KEY__as_attachment] is not None:
                raise HttpStatusException(ERR__as_attachment_unexpected)
            response.response_code = HTTPStatus.CREATED
            return self.url + "/" + DIR__render_template + "/" + res[KEY__document_id] + "." + res[KEY__render_as]
        else:
            return self.url + "/api/rendered_documents/" + res[KEY__document_id]

    def fetch_document_stream(self, inputs: dict):
        res = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_rendered_document,
                                                   {KEY__document_id: inputs[KEY__document_id]}, singleton_message=ERR__document_id_not_found,
                                                   as_objects=True)

        if not res[KEY__completed]:
            raise HttpStatusException(ERR__document_still_rendering, HTTP_STATUS__too_early)

        if res[KEY__create_file]:
            raise HttpStatusException(ERR__document_created_file)
        else:
            content = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__purge_rendered_document,
                                                           {KEY__document_id: inputs[KEY__document_id]}, as_objects=True)[KEY__content]
            as_attachment = False
            if inputs[KEY__as_attachment]:
                as_attachment = True

            buffer = BytesIO()
            buffer.write(content)
            buffer.seek(0)

            return send_file(buffer, as_attachment=as_attachment, download_name=res[KEY__filename])

    def render_document(self, inputs: dict, auth_token: str, ip_address: str):
        inputs[KEY__oauth_token] = self.refresh_auth_token(auth_token, ip_address)
        return execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__ins_rendered_document, inputs,
                                                    encryption_key=self.get_db_crypt_key(), encrypt_parameters=[KEY__parameters, KEY__oauth_token],
                                                    as_objects=True)

    def send_email(self, inputs: dict, auth_token: str, user_id: str, ip_address: str, username: str):
        if inputs[KEY__recipient] is not None:
            raise HttpStatusException(HTTPStatus.NOT_IMPLEMENTED.description, HTTPStatus.NOT_IMPLEMENTED.value)

        if inputs[KEY__recipient] is not None and inputs[KEY__attachments] is not None:
            raise HttpStatusException(ERR__cant_send_attachments)

        template = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_email_template_with_app_url, {
            KEY__name: inputs[KEY__email_template],
            KEY__configuration: inputs[KEY__configuration],
            KEY__application: inputs[KEY__application]
        }, as_objects=True)
        app_url = template.pop(KEY__application_url)
        artifact_url = template.pop(KEY__artifact_base_uri)

        params = inputs[KEY__parameters]
        if params is not None and template[KEY__data_validation_table] is None:
            raise HttpStatusException(ERR__unexpected_parameters)
        if template[KEY__data_validation_table] is not None and params is None:
            params = {}
        if template[KEY__data_validation_table] is not None:
            if EMAIL_PARAM__app_url in params:
                raise HttpStatusException(ERR__unexpected_validation_column % EMAIL_PARAM__app_url)

            interface = self.fetch_interface(template[KEY__application], inputs[KEY__configuration], template[KEY__schema])
            params, _ = self.fetch_sanitized_email_params(interface, template, params)

        optional_params = {EMAIL_PARAM__app_url: app_url}
        optional_params = {**optional_params, **params}

        recipients = inputs[KEY__recipient]
        if recipients is None:
            recipients = username

        if inputs[KEY__attachments] is not None:
            for attachment in inputs[KEY__attachments]:
                attachment[KEY__application] = inputs[KEY__application]
                attachment[KEY__configuration] = inputs[KEY__configuration]

        self.email_manager.construct_and_send_email(artifact_url, template, user_id, recipients, None, {}, optional_params,
                                                    attachments=EmailAttachment.deserialize_list(inputs[KEY__attachments], template[KEY__name],
                                                                                                 app_url),
                                                    attachment_access_token=self.refresh_auth_token(auth_token, ip_address))

    def reset_password_perform_reset(self, inputs: dict):
        token = inputs[KEY__reset_key]
        password = inputs[KEY__password]

        resp = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__reset_fetch, parameters={KEY__reset_or_poll_key: token},
                                                    decrypt_columns=[KEY__email], encryption_key=self.get_db_crypt_key(),
                                                    as_objects=True, singleton_message=ERR__cant_find_sign_up)
        self.add_account_password(str(resp[ATTR__the_user]), password)
        execute_supplied_statement(self.jaaql_lookup_connection, QUERY__reset_close, parameters={KEY__reset_key: token})

        res = {
            KEY__email: resp[KEY__email]
        }
        if resp[KEY__email_template] and resp[ATTR__data_lookup_json]:
            template = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_email_template_with_app_url,
                                                            {KEY__name: resp[KEY__email_template],
                                                             KEY__application: resp[KEY__application], KEY__configuration: resp[KEY__configuration]},
                                                            as_objects=True)
            interface = self.fetch_interface(template[KEY__application], resp[KEY__configuration], template[KEY__schema])
            res[KEY__parameters] = self.select_from_data_validation_table(interface, template[KEY__data_validation_table],
                                                                          json.loads(resp[ATTR__data_lookup_json]))

        return res

    def fetch_public_user(self, application: str, configuration: str):
        return execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_public_account_credentials,
                                                    {KEY__application: application, KEY__configuration: configuration},
                                                    as_objects=True, encryption_key=self.get_db_crypt_key(),
                                                    decrypt_columns=[KEY__password, KEY__username])

    def create_account(self, connection: DBInterface, inputs: dict):
        if inputs.get(KEY__password) is not None:
            crypt_utils.validate_password(inputs[KEY__password])

        if KEY__application not in inputs:
            inputs[KEY__application] = None
        if KEY__configuration not in inputs:
            inputs[KEY__configuration] = None
        if KEY__password not in inputs:
            inputs[KEY__password] = None
        if inputs[KEY__password] is None and inputs[KEY__application] is not None:
            inputs[KEY__password] = str(uuid.uuid4())  # Set a default password for public user
        the_password = inputs[KEY__password]
        query = QUERY__create_account
        account_id = execute_supplied_statement_singleton(connection, query, inputs,
                                                          encryption_key=self.get_db_crypt_key(), encrypt_parameters=[KEY__username, KEY__password],
                                                          encryption_salts={KEY__username: self.get_repeatable_salt()})[RET__rows][0]

        if the_password:
            self.add_account_password(account_id, the_password)

        return account_id

    def is_installed(self, response: JAAQLResponse):
        if not self.has_installed:
            response.response_code = HTTPStatus.UNPROCESSABLE_ENTITY
            return ERR__not_yet_installed

    def is_alive(self):
        version = execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__postgres_version, as_objects=True)[ATTR__version]
        if "PostgreSQL " not in version:
            raise HttpStatusException(ERR__keep_alive_failed)

    def install(self, db_connection_string: str, jaaql_password: str, super_db_password: str, install_key: str, allow_uninstall: bool):
        if self.jaaql_lookup_connection is None:
            if allow_uninstall:
                self.vault.insert_obj(VAULT_KEY__allow_jaaql_uninstall, True)

            if install_key != self.install_key:
                raise HttpStatusException(ERR__incorrect_install_key, HTTPStatus.UNAUTHORIZED)

            if db_connection_string is None:
                db_connection_string = PG__default_connection_string % os.environ[PG_ENV__password]

            address, port, _, username, db_password = DBInterface.fracture_uri(db_connection_string)

            super_interface = create_interface(self.config, address, port, DB__jaaql, username, db_password)
            conn = super_interface.get_conn()
            super_interface.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts, "install_1.sql"))
            resp = super_interface.execute_query(conn, QUERY__setup_jaaql_role)
            jaaql_db_password = resp[1][0][0]
            super_interface.commit(conn)
            super_interface.put_conn(conn)
            super_interface.close()

            self.jaaql_lookup_connection = create_interface(self.config, address, port, DB__jaaql, ROLE__jaaql, jaaql_db_password)
            conn = self.jaaql_lookup_connection.get_conn()
            self.jaaql_lookup_connection.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts, "install_2.sql"))
            self.jaaql_lookup_connection.commit(conn)
            self.jaaql_lookup_connection.put_conn(conn)

            execute_supplied_statement(self.jaaql_lookup_connection, QUERY__setup_system)

            execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__attach_account, {
                KEY__user_id: ROLE__jaaql,
                KEY__enc_username: USERNAME__jaaql
            }, encryption_key=self.get_db_crypt_key(), encrypt_parameters=[KEY__enc_username],
                                                 encryption_salts={KEY__enc_username: self.get_repeatable_salt()})
            execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__attach_account, {
                KEY__user_id: ROLE__postgres,
                KEY__enc_username: USERNAME__super_db
            }, encryption_key=self.get_db_crypt_key(), encrypt_parameters=[KEY__enc_username],
                                                 encryption_salts={KEY__enc_username: self.get_repeatable_salt()})

            self.add_account_password(ROLE__postgres, super_db_password)
            self.add_account_password(ROLE__jaaql, jaaql_password)
            self.jaaql_lookup_connection.close()

            super_conn_str = PROTOCOL__postgres + username + ":" + db_password + "@" + address + ":" + str(port) + "/"
            self.vault.insert_obj(VAULT_KEY__super_db_credentials, super_conn_str)
            jaaql_lookup_str = PROTOCOL__postgres + ROLE__jaaql + ":" + jaaql_db_password + "@" + address + ":" + str(port) + "/" + DB__jaaql
            self.vault.insert_obj(VAULT_KEY__jaaql_lookup_connection, jaaql_lookup_str)

            print("Rebooting to allow JAAQL config to be shared among workers")
            threading.Thread(target=self.exit_jaaql).start()
        else:
            raise HttpStatusException(ERR__already_installed)

    def verification_thread(self, the_queue: Queue):
        print("Starting auth verification thread")
        while True:
            auth_token, ip_address, complete = the_queue.get()
            try:
                self.verify_auth_token(auth_token, ip_address)
                complete.put((True, None, None))
            except HttpStatusException as ex:
                complete.put((False, ex.message, ex.response_code))
            except Exception as ex:
                complete.put((False, str(ex), 500))

    def verify_auth_token_threaded(self, auth_token: str, ip_address: str, complete: Queue):
        try:
            if JAAQLModel.VERIFICATION_QUEUE is None:
                JAAQLModel.VERIFICATION_QUEUE = Queue()
                threading.Thread(target=self.verification_thread, args=[JAAQLModel.VERIFICATION_QUEUE], daemon=True).start()
            JAAQLModel.VERIFICATION_QUEUE.put((auth_token, ip_address, complete))
            payload = json.loads(base64url_decode(auth_token.split(".")[1].encode("UTF-8")).decode())
            return payload[KEY__user_id], payload[KEY__username], payload[KEY__ip_id], payload[KEY__is_public]
        except Exception:
            raise HttpStatusException(ERR__invalid_token, HTTPStatus.UNAUTHORIZED)

    def verify_auth_token(self, auth_token: str, ip_address: str):
        decoded = crypt_utils.jwt_decode(self.vault.get_obj(VAULT_KEY__jwt_crypt_key), auth_token, JWT_PURPOSE__oauth)
        if not decoded or decoded[KEY__address] != ip_address:
            raise HttpStatusException(ERR__invalid_token, HTTPStatus.UNAUTHORIZED)
        lookup = {KEY__user_id: decoded[KEY__user_id], KEY__password: decoded[KEY__password_id]}
        execute_supplied_statement_singleton(self.jaaql_lookup_connection, QUERY__fetch_matching_recent_passwords_with_account, lookup,
                                             singleton_message=ERR__invalid_token, singleton_code=HTTPStatus.UNAUTHORIZED, skip_commit=True)

        return decoded[KEY__user_id], decoded[KEY__username], decoded[KEY__ip_id], decoded[KEY__is_public]

    def refresh_cached_canned_query_service(self, connection: DBInterface, application: str, configuration: str):
        self.check_is_internal_admin(connection)

        self.cached_canned_query_service.refresh_configuration(self.is_container, self.jaaql_lookup_connection, application, configuration)

    def refresh_auth_token(self, auth_token: str, ip_address: str):
        decoded = crypt_utils.jwt_decode(self.vault.get_obj(VAULT_KEY__jwt_crypt_key), auth_token, JWT_PURPOSE__oauth, allow_expired=True)
        if not decoded:
            raise HttpStatusException(ERR__invalid_token, HTTPStatus.UNAUTHORIZED)

        if datetime.fromisoformat(decoded[KEY__created]) + timedelta(milliseconds=self.refresh_expiry_ms) < datetime.now():
            raise HttpStatusException(ERR__refresh_expired, HTTPStatus.UNAUTHORIZED)

        return self.get_auth_token(decoded[KEY__username], ip_address)

    def get_auth_token(self, username: str, ip_address: str, password: str = None, response: JAAQLResponse = None):
        incorrect_credentials = False
        user_id = None
        res = None
        encrypted_ip = None

        try:
            user_id = self.fetch_user_from_username(username, singleton_code=HTTPStatus.UNAUTHORIZED)
        except HttpStatusException as exc:
            if exc.response_code == HTTPStatus.UNAUTHORIZED:
                incorrect_credentials = True
            else:
                raise exc

        if not incorrect_credentials:
            salt_user = self.get_repeatable_salt(user_id)

            encrypted_ip = jaaql__encrypt(ip_address, self.get_db_crypt_key(), salt_user)

            the_kwargs = {
                "encryption_salts": {KEY__enc_username: self.get_repeatable_salt()},
                "decrypt_columns": [KEY__password_hash, KEY__username],
                "encrypt_parameters": [KEY__enc_username],
                "encryption_key": self.get_db_crypt_key(),
                "as_objects": True
            }

            res = execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fetch_recent_passwords_with_ips, {
                KEY__enc_username: username,
                KEY__encrypted_address: encrypted_ip,
            }, **the_kwargs)

            if len(res) == 0 and password is not None:
                res = execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fetch_recent_passwords, {
                    KEY__enc_username: username
                }, **the_kwargs)

            if len(res) == 1:
                res = res[0]
                if password is not None and not crypt_utils.verify_password_hash(res[KEY__password_hash], password,
                                                                                 salt=self.get_repeatable_salt(res[KEY__user_id])):
                    incorrect_credentials = True
            else:
                incorrect_credentials = True

        if incorrect_credentials:
            raise HttpStatusException(ERR__incorrect_credentials, HTTPStatus.UNAUTHORIZED)

        address = execute_supplied_statement_singleton(self.jaaql_lookup_connection,
                                                       QUERY__add_or_update_ip,
                                                       {KEY__account: res[KEY__user_id], KEY__encrypted_address: encrypted_ip},
                                                       as_objects=True)[KEY__address]

        jwt_data = {
            KEY__user_id: str(res[KEY__user_id]),
            KEY__password_id: str(res[KEY__password_id]),
            KEY__username: res[KEY__username],
            KEY__address: ip_address,
            KEY__ip_id: str(address),
            KEY__created: datetime.now().isoformat(),
            KEY__is_public: res[KEY__application] is not None
        }

        if response is not None:
            response.user_id = str(res[KEY__user_id]),
            response.ip_id = str(address)

        return crypt_utils.jwt_encode(self.vault.get_obj(VAULT_KEY__jwt_crypt_key), jwt_data, JWT_PURPOSE__oauth, expiry_ms=self.token_expiry_ms)

    def fetch_pivoted_application_schemas(self):
        data = execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fetch_application_schemas, as_objects=True,
                                          skip_commit=True)

        return self.group(self.pivot(data,
                          [JAAQLPivotInfo(KEY__configuration, "name", "configuration_name"),
                           JAAQLPivotInfo(KEY__schema, "configuration_name", "schema_name")]), KEY__name)

    def create_interface_for_db(self, user_id: str, database: str, sub_role: str = None):
        jaaql_uri = self.vault.get_obj(VAULT_KEY__super_db_credentials)
        address, port, _, username, password = DBInterface.fracture_uri(jaaql_uri)
        return create_interface(self.config, address, port, database, username, password=password, role=user_id, sub_role=sub_role)

    def submit(self, inputs: dict, user_id: str, verification_hook: Queue):
        if not isinstance(inputs, dict):
            raise HttpStatusException("Expected object or string input")

        if KEY__application in inputs:
            application_configs = self.fetch_pivoted_application_schemas()
            app = application_configs[inputs[KEY__application]]
            config = app[inputs[KEY__configuration]]

            found_db = None
            if KEY__schema in inputs:
                found_db = config[inputs[KEY__schema]][KEY__database]
                inputs.pop(KEY__schema)
            else:
                if len(config) == 1:
                    found_db = config[list(config.keys())[0]][KEY__database]
                else:
                    found_dbs = [val[KEY__database] for _, val in config.items() if val[KEY__is_default]]
                    if len(found_dbs) == 1:
                        found_db = found_dbs[0]

            if not found_db:
                raise HttpStatusException(ERR__schema_invalid)

            inputs[KEY__database] = found_db

        if KEY__database not in inputs:
            inputs[KEY__database] = DB__jaaql

        sub_role = inputs.pop(KEY__role) if KEY__role in inputs else None
        required_db = self.create_interface_for_db(user_id, inputs[KEY__database], sub_role)

        return InterpretJAAQL(required_db).transform(inputs, skip_commit=inputs.get(KEY__read_only), wait_hook=verification_hook,
                                                     encryption_key=self.get_db_crypt_key(),
                                                     canned_query_service=self.cached_canned_query_service)



from jaaql.patch import monkey_patch

monkey_patch()

from jaaql.email.email_manager_service import create_flask_app
import os
from jaaql.constants import ENVIRON__vault_key, ENVIRON__email_credentials

create_flask_app(os.environ.get(ENVIRON__vault_key), os.environ.get(ENVIRON__email_credentials), True)



from http import HTTPStatus
from argparse import Namespace

RESP__default_err_message = "We have encountered an error whilst processing your request!"
RESP__default_err_code = HTTPStatus.INTERNAL_SERVER_ERROR

HTTP_STATUS_CONNECTION_EXPIRED = Namespace(value=419)
ERR__connection_expired = "Connection expired"

ERR__already_installed = "JAAQL has already been installed!"
ERR__already_signed_up = "User has already signed up!"
ERR__non_node_connection_object = "Cannot request a list of databases for a non-node connection object"
ERR__passwords_do_not_match = "The supplied passwords do not match!"
ERR__cannot_override_db = "Cannot override DB"


class HttpStatusException(Exception):

    def __init__(self, message: str, response_code: int = HTTPStatus.UNPROCESSABLE_ENTITY):
        super().__init__(message)

        if response_code is None:
            response_code = HTTPStatus.UNPROCESSABLE_ENTITY

        self.message = message
        self.response_code = response_code



from logging import StreamHandler
from jaaql.constants import ENVIRON__vault_key, ENVIRON__jaaql_profiling
import re
import logging
import pkgutil
from jaaql.openapi.swagger_documentation import produce_all_documentation
from jaaql.email.email_manager_service import create_flask_app as create_email_service_app
from jaaql.services.script_install import bootup
from typing import List
from jaaql.utilities.options import parse_options, OPT_KEY__vault_key, OPT_KEY__email_credentials, OPT_KEY__local_install, Option, OPT_KEY__profiling
import sys
import threading
import os
import jaaql.documentation as documentation
from jaaql.utilities.utils import load_config, get_base_url
from jaaql.mvc.controller import JAAQLController
from jaaql.mvc.model import JAAQLModel
from jaaql.mvc.controller_interface import JAAQLControllerInterface
from jaaql.mvc.model_interface import JAAQLModelInterface
from jaaql.config_constants import *
import inspect
from os.path import join, dirname, basename


DEFAULT__mfa_label = "test"

WARNING__vault_key_stdin = "MAJOR SECURITY ISSUE! Passing vault key via program arguments is insecure as other progra" \
                           "ms can see the arguments. Please provide via stdin instead!"

WARNING__audit_off = "Audit trail is off. Logs will still be kept by the internal postgres instance"


class SensitiveHandler(StreamHandler):

    PATTERN = r"(?:\?|&|;)([^=]+)=([^\s&|;]+)"

    def emit(self, record):
        msg = self.format(record)
        msg = re.sub(self.PATTERN, lambda x: x[0][0:len(x[0]) - len(x[2])] + min(len(x[2]), 5) * '*', msg)
        if "GET /internal/is_installed HTTP/1.1" not in msg:
            print(msg)


def dir_non_builtins(folder):
    if isinstance(folder, str):
        pkgpath = folder
    else:
        pkgpath = os.path.dirname(folder.__file__)
    pre_dirs = [(loader, name) for loader, name, _ in pkgutil.walk_packages([pkgpath])]
    dirs = []
    for sub_dir in pre_dirs:
        if sub_dir[1] in dir(folder):
            dirs.append(getattr(folder, sub_dir[1]))
        else:
            dirs.append(sub_dir[0].find_module(sub_dir[1]).load_module(sub_dir[1]))
    return [doc for doc in dirs if doc.__name__ != "builtins"]


def create_app(override_config_path: str = None, controllers: [JAAQLControllerInterface] = None, models: [JAAQLModelInterface] = None,
               additional_options: List[Option] = None, supplied_documentation = None, **kwargs):
    is_gunicorn = os.environ.get("JAAQL_DEPLOYED") == "true"

    if controllers is None:
        controllers = []
    if models is None:
        models = []
    if not isinstance(controllers, list):
        controllers = [controllers]
    if not isinstance(models, list):
        models = [models]

    if supplied_documentation is None:
        supplied_documentation = []
    if not isinstance(supplied_documentation, list):
        supplied_documentation = [supplied_documentation]

    if not is_gunicorn:
        options = parse_options(sys.argv, False, additional_options=additional_options)
    else:
        options = {}

    if not is_gunicorn:
        threading.Thread(target=create_email_service_app, args=[options.get(OPT_KEY__vault_key), options.get(OPT_KEY__email_credentials)],
                         daemon=True).start()
        threading.Thread(target=bootup, args=[options.get(OPT_KEY__vault_key), False, OPT_KEY__local_install in options], daemon=True).start()

    if OPT_KEY__vault_key in options:
        print(WARNING__vault_key_stdin, file=sys.stderr)
        vault_key = options[OPT_KEY__vault_key]
    elif is_gunicorn:
        vault_key = os.environ.get(ENVIRON__vault_key)
    else:
        vault_key = input("Input vault key: ")

    logging.getLogger("werkzeug").handlers.append(SensitiveHandler())

    config = load_config(is_gunicorn, override_config_path)

    mfa_name = config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__mfa_label]
    if mfa_name == DEFAULT__mfa_label:
        print("MFA label is set to default. This isn't a security issue but adds a nice name when added to "
              "authenticator apps via QR codes. You can change in the config")

    force_mfa = config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__force_mfa] in ("true", "True", True, "TRUE")
    do_audit = config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__do_audit] in ("true", "True", True, "TRUE")

    # The following code sets it "properly" in case it is accessed incorrectly from elsewhere
    config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__force_mfa] = force_mfa
    config[CONFIG_KEY__security][CONFIG_KEY_SECURITY__do_audit] = do_audit

    if not do_audit:
        print(WARNING__audit_off, file=sys.stderr)

    url = config[CONFIG_KEY__swagger][CONFIG_KEY_SWAGGER__url]

    migration_folder = join(dirname(inspect.stack()[1].filename), "migrations")
    migration_project_name = basename(dirname(inspect.stack()[1].filename))
    if not os.path.exists(migration_folder):
        migration_folder = None
        migration_project_name = None

    model = JAAQLModel(config, vault_key, options, migration_project_name, migration_folder, is_container=is_gunicorn, url=url)
    controller = JAAQLController(model, is_gunicorn, get_base_url(config, is_gunicorn),
                                 OPT_KEY__profiling in options or os.environ.get(ENVIRON__jaaql_profiling))
    controller.create_app()

    for sub_controller in controllers:
        sub_controller.route(controller)

    for sub_model in models:
        sub_model.set_model(model)

    doc_modules = dir_non_builtins(documentation)
    all_docs = []
    for doc in supplied_documentation:
        all_docs.extend(dir_non_builtins(doc))

    base_path = None
    if is_gunicorn:
        base_path = "www"
        url += "/api/"

    produce_all_documentation(doc_modules + all_docs, url, is_prod=is_gunicorn, base_path=base_path)

    if is_gunicorn:
        return controller.app
    else:
        controller.app.run(port=int(config[CONFIG_KEY__server][CONFIG_KEY_SERVER__port]), host="0.0.0.0", threaded=True)



from jaaql.openapi.swagger_documentation import *
from jaaql.constants import *
from jaaql.documentation.documentation_shared import rename_arg, ARG_RES__application, ARG_RES__configuration

TITLE = "JAAQL Internal API"
DESCRIPTION = "Collection of methods in the JAAQL internal API"
FILENAME = "jaaql_internal_api"

ARG_RES__install_key = SwaggerArgumentResponse(
    name=KEY__install_key,
    description="Single use/purpose key. Find in the logs, line starting with 'INSTALL KEY:' at startup",
    arg_type=str,
    example=["aefc1b08-a573-466f-bdd0-706ae281cc99", "ec63aa9a-b189-419f-8e0b-7fcc4ff8c857"],
    required=True
)

ARG_RES__uninstall_key = rename_arg(ARG_RES__install_key, KEY__uninstall_key, ARG_RES__install_key.description.replace("INSTALL", "UNINSTALL"))

DOCUMENTATION__canned_queries = SwaggerDocumentation(
    tags="Application",
    methods=SwaggerMethod(
        name="Refresh Canned Queries",
        description="Refreshes the application configuration canned queries.",
        method=REST__POST,
        arguments=[
            ARG_RES__application,
            ARG_RES__configuration
        ]
    )
)

DOCUMENTATION__install = SwaggerDocumentation(
    tags="Installation",
    security=False,  # This method is not secured as the system is not setup yet. It uses a OTP token system via logs
    methods=SwaggerMethod(
        name="Install JAAQL",
        description="Installs JAAQL to the configured database. Please allow a minute after running for the server to "
                    "refresh and reload. A connection string is required for local installation and is not allowed when running "
                    "inside a docker container as the docker container comes with a postgres database and JAAQL will use that",
        method=REST__POST,
        body=[
            SwaggerArgumentResponse(
                name="db_connection_string",
                description="Database connection string for the superuser",
                arg_type=str,
                example=["postgresql://postgres:123456@localhost:5432/",
                         "postgresql://postgres:pa55word@localhost:5432/"],
                required=True,
                local_only=True
            ),
            ARG_RES__install_key,
            SwaggerArgumentResponse(
                name=KEY__jaaql_password,
                description="The password for the jaaql user, can create applications, configurations",
                example=["pa55word"],
                strip=True,
                arg_type=str
            ),
            SwaggerArgumentResponse(
                name=KEY__super_db_password,
                description="Gives access to the database super user",
                example=["passw0rd"],
                required=True,
                strip=True,
                arg_type=str
            ),
            SwaggerArgumentResponse(
                name=KEY__allow_uninstall,
                description="Allows uninstall. DO NOT USE ON PROD, used to aid with testing. Defaults to false if not supplied",
                example=[False, True],
                required=False,
                condition="If you want to specify if uninstallation allowed",
                arg_type=bool
            )
        ],
        response=[
            SwaggerFlatResponse(),
            SwaggerFlatResponse(
                description=ERR__already_installed,
                code=HTTPStatus.UNPROCESSABLE_ENTITY,
                body=ERR__already_installed
            )
        ]
    )
)

DOCUMENTATION__uninstall = SwaggerDocumentation(
    tags="Installation",
    security=False,  # Would not matter which user uses it, needs root db credentials
    methods=SwaggerMethod(
        name="Uninstall JAAQL",
        description="Uninstalls the JAAQL system",
        method=REST__POST,
        body=[
            SwaggerArgumentResponse(
                name=KEY__db_super_user_password,
                description="The password for the database super user account",
                example=["123456"],
                arg_type=str
            ),
            ARG_RES__uninstall_key
        ]
    )
)

DOCUMENTATION__is_installed = SwaggerDocumentation(
    tags="Installation",
    security=False,
    methods=SwaggerMethod(
        name="Is installed",
        description="Returns 200 OK if the service has installed otherwise a 422",
        method=REST__GET,
        response=[
            SwaggerFlatResponse(),
            SwaggerFlatResponse(
                description=ERR__not_yet_installed,
                code=HTTPStatus.UNPROCESSABLE_ENTITY,
                body=ERR__not_yet_installed
            )
        ]
    )
)

# Not unused. Used to generate html files
from jaaql.documentation.documentation_shared import DOCUMENTATION__oauth_token, DOCUMENTATION__oauth_refresh

DOCUMENTATION__is_alive = SwaggerDocumentation(
    security=False,
    tags="Is Alive",
    methods=SwaggerMethod(
        name="Check is alive",
        description="An endpoint that can be called by any service to see if the service is alive",
        method=REST__GET
    )
)

DOCUMENTATION__deploy = SwaggerDocumentation(
    tags="Deployment",
    methods=[
        SwaggerMethod(
            name="Redeploys the JAAQL service",
            description="Will redeploy the JAAQL service, loading the latest version from git. In the future will "
                        "support loading a specific tagged version",
            method=REST__POST
        )
    ]
)



import requests
from datetime import datetime
import threading
import random

API_URL = "http://localhost/api"
ENDPOINT_OAUTH = "/oauth/token"
ENDPOINT_SUBMIT = "/submit"
ENDPOINT_CONFIGS = "/configurations"
ENDPOINT_ARGS = "/configurations/arguments"
CONCURRENT = 1
CREDS = [
    {
        "username": "jaaql",
        "password": "pa55word"
    },
    {
        "username": "superjaaql",
        "password": "passw0rd"
    }
]


def run_perf_tests():
    while True:
        # res = requests.post(API_URL + ENDPOINT_OAUTH, json=CREDS[1 if random.random() > 0.5 else 0])
        res = requests.post(API_URL + ENDPOINT_OAUTH, json=CREDS[0])
        print(datetime.now(), end=" ")
        print(res.status_code)
        if res.status_code != 200:
            continue
        token = res.json()

        auth_headers = {
            "Authentication-Token": token
        }

        res = requests.get(API_URL + ENDPOINT_CONFIGS + "?application=console", headers=auth_headers)
        print(datetime.now(), end=" ")
        print(res.status_code)
        if res.status_code != 200:
            continue
        configuration_name = res.json()[0]["configuration"]

        res = requests.get(API_URL + ENDPOINT_ARGS + "?application=console&configuration=" + configuration_name,
                           headers=auth_headers)
        print(datetime.now(), end=" ")
        print(res.status_code)
        if res.status_code != 200:
            continue
        connection = res.json()[0]["connection"]

        res = requests.post(API_URL + ENDPOINT_SUBMIT, json={
            "query": "SELECT * FROM current_database();",
            "connection": connection,
            "database": "database" + str(random.randint(0, 9))
        }, headers=auth_headers)

        print(datetime.now(), end=" ")
        print(res.status_code)


if __name__ == "__main__":
    for i in range(CONCURRENT):
        threading.Thread(target=run_perf_tests).start()


from datetime import datetime
import requests
from jaaql.utilities.utils_no_project_imports import time_delta_ms
import time

MAX_STARTUP_TIME_MS = 120000
BASE_URL = "http://127.0.0.1/api"


def wait_for_service():
    service_started = False
    start_time = datetime.now()
    finish_time = None
    while not service_started:
        try:
            print("Awaiting CTP bootup")
            status = requests.get(BASE_URL + "/ctp/hello").status_code
            if status == 200:
                service_started = True
                finish_time = datetime.now()
        except:
            pass
        if time_delta_ms(start_time, datetime.now()) > MAX_STARTUP_TIME_MS:
            raise Exception("CTP did not boot in the required time")
        time.sleep(5)

    print("CTP booted in " + str(time_delta_ms(start_time, finish_time)) + "ms")



from jaaql.mvc.model import JAAQLModel
from jaaql.mvc.base_controller import BaseJAAQLController
from jaaql.documentation.documentation_internal import *
from jaaql.documentation.documentation_public import *
from jaaql.documentation.documentation_shared import *
from jaaql.mvc.response import JAAQLResponse
from jaaql.db.db_interface import DBInterface
import queue


class JAAQLController(BaseJAAQLController):

    def __init__(self, model: JAAQLModel, is_prod: bool, base_url: str, do_profiling: bool = False):
        super().__init__(model, is_prod, base_url, do_profiling)

    def create_app(self):

        @self.cors_route('/oauth/token', DOCUMENTATION__oauth_token)
        def fetch_oauth_token(http_inputs: dict, ip_address: str, response: JAAQLResponse):
            return self.model.get_auth_token(**http_inputs, ip_address=ip_address, response=response)

        @self.cors_route('/internal/redeploy', DOCUMENTATION__deploy)
        def redeploy(connection: DBInterface):
            return self.model.redeploy(connection)

        @self.cors_route(ENDPOINT__refresh, DOCUMENTATION__oauth_refresh)
        def refresh_oauth_token(auth_token: str, ip_address: str):
            return self.model.refresh_auth_token(auth_token, ip_address)

        @self.cors_route('/internal/canned-queries', DOCUMENTATION__canned_queries)
        def refresh_app_config(connection: DBInterface, inputs: dict):
            self.model.refresh_cached_canned_query_service(connection, **inputs)

        @self.cors_route(ENDPOINT__install, DOCUMENTATION__install)
        def install(http_inputs: dict):
            return self.model.install(**http_inputs)

        @self.cors_route('/internal/uninstall', DOCUMENTATION__uninstall)
        def uninstall():
            raise HttpStatusException(HTTPStatus.NOT_IMPLEMENTED.description, HTTPStatus.NOT_IMPLEMENTED.value)

        @self.cors_route('/internal/is_installed', DOCUMENTATION__is_installed)
        def is_installed(response: JAAQLResponse):
            return self.model.is_installed(response)

        @self.cors_route(ENDPOINT__is_alive, DOCUMENTATION__is_alive)
        def is_alive():
            self.model.is_alive()

        @self.cors_route('/email/account', DOCUMENTATION__drop_email_account)
        def drop_email_account(http_inputs: dict, connection: DBInterface):
            self.model.drop_email_account(connection, **http_inputs)

        @self.cors_route('/accounts', DOCUMENTATION__create_account)
        def accounts(connection: DBInterface, http_inputs: dict):
            self.model.create_account(connection, http_inputs)

        @self.cors_route('/account/password', DOCUMENTATION__password)
        def password(user_id: str, username: str, ip_address: str, http_inputs: dict):
            return self.model.add_my_account_password(user_id, username, ip_address, **http_inputs)

        @self.cors_route('/public', DOCUMENTATION__public_user)
        def public_user(http_inputs: dict):
            return self.model.fetch_public_user(**http_inputs)

        @self.cors_route('/submit', DOCUMENTATION__submit)
        def submit(http_inputs: dict, user_id: str, verification_hook: queue.Queue):
            return self.model.submit(http_inputs, user_id, verification_hook=verification_hook)

        @self.cors_route('/account/signup/request', DOCUMENTATION__sign_up_request_invite)
        def signup_request(http_inputs: dict):
            return self.model.request_signup(http_inputs)

        @self.cors_route('/account/signup/status', DOCUMENTATION__sign_up_status)
        def signup_request(http_inputs: dict):
            return self.model.signup_status(http_inputs)

        @self.cors_route('/account/signup/activate', DOCUMENTATION__sign_up_with_invite)
        def signup_activate(http_inputs: dict):
            return self.model.sign_up_user_with_token(http_inputs[KEY__invite_key], http_inputs[KEY__password])

        @self.cors_route('/account/signup/fetch', DOCUMENTATION__sign_up_fetch)
        def signup_fetch(http_inputs: dict):
            return self.model.fetch_signup(http_inputs[KEY__invite_key])

        @self.cors_route('/account/signup/finish', DOCUMENTATION__sign_up_finish)
        def signup_finish(http_inputs: dict):
            self.model.finish_signup(http_inputs[KEY__invite_key])

        @self.cors_route('/account/reset-password', DOCUMENTATION__reset_password)
        def reset_password(http_inputs: dict):
            return self.model.send_reset_password_email(http_inputs)

        @self.cors_route('/account/reset-password/status', DOCUMENTATION__reset_password_status)
        def reset_password_status(http_inputs: dict):
            return self.model.reset_password_status(http_inputs)

        @self.cors_route('/account/reset-password/reset', DOCUMENTATION__reset_password_with_invite)
        def reset_password_reset(http_inputs: dict):
            return self.model.reset_password_perform_reset(http_inputs)

        @self.cors_route(ENDPOINT__jaaql_emails, DOCUMENTATION__email)
        def emails(http_inputs: dict, auth_token: str, user_id: str, ip_address: str, is_public: bool, username: str):
            if is_public:
                raise HttpStatusException(ERR__user_public, HTTPStatus.UNAUTHORIZED)

            return self.model.send_email(http_inputs, auth_token, user_id, ip_address, username)

        @self.cors_route('/documents', DOCUMENTATION__document)
        def documents(http_inputs: dict, auth_token: str, ip_address: str, response: JAAQLResponse):
            if self.is_get():
                return self.model.fetch_document(http_inputs, response)
            else:
                return self.model.render_document(http_inputs, auth_token, ip_address)

        @self.cors_route('/rendered_documents', DOCUMENTATION__rendered_document)
        def documents(http_inputs: dict):
            return self.model.fetch_document_stream(http_inputs)



from typing import List, Optional, Union, Any, TypeVar
from types import ModuleType
from jaaql.exceptions.http_status_exception import *
from jaaql.constants import EXAMPLE__jwt, HEADER__security
import os
import shutil
import yaml as yaml_utils
import json
from jaaql.utilities.utils import get_jaaql_root
from jaaql.constants import VERSION

ERR__empty_example_list = "Empty examples list for argument '%s'"
ERR__null_example = "No examples found for argument '%s'"
ERR__example_type = "Found type '%s' for example #%d, argument '%s'. Doesn't match type '%s'"
ERR__response_type = "Flat response with http code #%d, argument '%s'. Doesn't match type '%s'"
ERR__duplicated_response_exception_code = "HTTP response error code #%d was duplicated"
ERR__duplicated_argument_response_name = "Field '%s' was duplicated in the argument or response"
ERR__duplicated_example = "Example was duplicated for '%s'"
ERR__duplicated_method = "HTTP method '%s' was duplicated"
ERR__condition_not_needed = "Condition supplied for '%s' but is argument is required. Please remove condition"
ERR__condition_needed = "Condition not supplied for '%s' but is argument is not required. Please supply condition"
ERR__documentation_variable_expected = "Expected '%s' variable in documentation module '%s'"
ERR__empty_path = "Empty path for documentation. Variable unused or documentation occuring before controller import"
ERR__nested_not_allowed = "Nested objects are not allowed in method arguments. Please place in the body"
ERR__list_not_allowed = "Lists now allowed in method arguments. Please place in the body"
ERR__delete_not_allowed_body = "HTTP method delete isn't allowed a body. Please use the arguments parameter instead"
ERR__get_not_allowed_body = "HTTP method get isn't allowed a body. Please use the arguments parameter instead"

TYPE__example = Union[List[Any], Any]

T = TypeVar('T')

ATTR__description = "DESCRIPTION"
ATTR__title = "TITLE"
ATTR__output = "OUTPUT"
ATTR__version = "VERSION"
ATTR__filename = "FILENAME"

PATH__empty = ""

VERSION__open_api = "3.0.3"

CONTENT_TYPE__json = "application/json"  # This isn't a python constant so it is defined here

TYPE__string = "string"

OPEN_API__typemap = {
    bool: "boolean",
    float: "number",
    int: "integer",
    str: TYPE__string
}

OPEN_API__array = "array"
OPEN_API__content = "content"
OPEN_API__description = "description"
OPEN_API__example = "example"
OPEN_API__in = "in"
OPEN_API__in_header = "header"
OPEN_API__in_path = "path"
OPEN_API__in_query = "query"
OPEN_API__info = "info"
OPEN_API__items = "items"
OPEN_API__name = "name"
OPEN_API__object = "object"
OPEN_API__openapi = "openapi"
OPEN_API__parameters = "parameters"
OPEN_API__paths = "paths"
OPEN_API__path_var_open = "{"
OPEN_API__path_var_close = "}"
OPEN_API__properties = "properties"
OPEN_API__request_body = "requestBody"
OPEN_API__required = "required"
OPEN_API__responses = "responses"
OPEN_API__schema = "schema"
OPEN_API__servers = "servers"
OPEN_API__spacer = "  "
OPEN_API__summary = "summary"
OPEN_API__tags = "tags"
OPEN_API__title = "title"
OPEN_API__type = "type"
OPEN_API__url = "url"
OPEN_API__version = "version"

YAML__array_close = "]"
YAML__array_open = "["
YAML__array_separator = ", "
YAML__double_quote = '"'
YAML__quote = "'"
YAML__list_separator = "- "
YAML__new_line = '\n'
YAML__separator = ":"

DIR__swagger = "swagger"
DIR__apps = "apps"
DIR__scripts = "scripts"
FILE__swagger_template = "swagger_template.html"
EXTENSION__html = ".html"
EXTENSION__json = ".json"
EXTENSION__yaml = ".yaml"

RESPONSE__OK = "OK"
RESPONSE__200_ok = "200 OK"

REST__DELETE = "DELETE"
REST__GET = "GET"
REST__OPTIONS = "OPTIONS"
REST__POST = "POST"
REST__PUT = "PUT"

TEMPLATE__url = "url: '%s'"
TEMPLATE__spec = "spec: "

DESCRIPTION__security = "A JWT token fetched from /oauth/token"

DOCUMENTATION__allow_all = "ALLOW_ALL"  # Set as the name to allow all input/output


def isinstance_union(arg, check: Union) -> bool:
    return any([isinstance(arg, cls) for cls in check.__args__])


def force_list(potential_list: Union[List[T], T]) -> List[T]:
    return potential_list if isinstance(potential_list, list) else ([] if potential_list is None else [potential_list])


class SwaggerException(Exception):

    def __init__(self, *args, **kwargs):
        super().__init__(args, kwargs)


class SwaggerResponseException:

    def __init__(self, code: HTTPStatus = RESP__default_err_code, message: str = RESP__default_err_message):
        self.code = code
        self.message = message


TYPE__exceptions = Optional[Union[List[SwaggerResponseException], SwaggerResponseException]]


class SwaggerSimpleList:

    def __init__(self, arg_type: type, description: str, example: Optional[TYPE__example] = None,
                 required: bool = False, condition: str = None):
        self.arg_type = arg_type
        self.description = description
        self.example = example
        self.required = required
        self.condition = condition

        if not required and condition is None:
            raise SwaggerException(ERR__condition_needed % "simple list")

        self._validate_example(arg_type, self.example)

    @staticmethod
    def _validate_example(arg_type: type, example: Optional[TYPE__example] = None):
        if example is None:
            raise SwaggerException(ERR__null_example % "simple list")
        if len(example) == 0:
            raise SwaggerException(ERR__empty_example_list % "simple list")

        for ex, ex_id in zip(example, range(len(example))):
            if not isinstance(ex, arg_type):
                raise SwaggerException(ERR__example_type % (str(type(ex)), ex_id + 1, "simple list", str(arg_type)))

        if len(example) != len(set(example)):
            raise SwaggerException(ERR__duplicated_example % "simple list")


class SwaggerArgumentResponse:

    def __init__(self, name: str, description: str,
                 arg_type: Union[type, List['SwaggerArgumentResponse'], 'SwaggerList', SwaggerSimpleList],
                 example: Optional[TYPE__example] = None, required: bool = True, condition: str = None,
                 local_only: bool = False, strip: bool = True, lower: bool = False):
        self.name = name
        self.description = description
        if arg_type == bool and example is None:
            example = [True, False]
        self.example = force_list(example)
        self.required = required
        self.arg_type = arg_type
        self.condition = condition
        self.local_only = local_only
        self.strip = strip
        self.lower = lower

        if not required and condition is None and name != DOCUMENTATION__allow_all:
            raise SwaggerException(ERR__condition_needed % name)
        if required and condition is not None:
            raise SwaggerException(ERR__condition_not_needed % name)

        arg_type_sub_list = isinstance(arg_type, list)
        if arg_type_sub_list:
            arg_type_sub_list = all([isinstance(x, SwaggerArgumentResponse) for x in arg_type])

        if not arg_type_sub_list and not isinstance_union(arg_type, TYPE__swagger_list_like) and not self.is_arg_all():
            self._validate_example(name, arg_type, self.example)

    @staticmethod
    def _validate_example(name: str, arg_type: type, example: Optional[TYPE__example] = None):
        if example is None:
            raise SwaggerException(ERR__null_example % name)
        if len(example) == 0:
            raise SwaggerException(ERR__empty_example_list % name)

        for ex, ex_id in zip(example, range(len(example))):
            if not isinstance(ex, arg_type):
                raise SwaggerException(ERR__example_type % (str(type(ex)), ex_id + 1, name, str(arg_type)))

        if len(example) != len(set(example)):
            raise SwaggerException(ERR__duplicated_example % name)

    def is_arg_all(self):
        if isinstance(self.arg_type, SwaggerArgumentResponse):
            if self.arg_type.name == DOCUMENTATION__allow_all:
                return True
        return self.name == DOCUMENTATION__allow_all


class SwaggerList:

    def __init__(self, *args: Union[SwaggerArgumentResponse, 'SwaggerList']):
        self.responses = list(args)


TYPE__swagger_list_like = Union[SwaggerSimpleList, SwaggerList]

ARG_RESP__allow_all = SwaggerArgumentResponse(DOCUMENTATION__allow_all, DOCUMENTATION__allow_all, str)
BODY__file = "JAAQL_IS_FILE"

TYPE__argument_response = Optional[Union[List[SwaggerArgumentResponse], SwaggerArgumentResponse, SwaggerList,
                                         SwaggerSimpleList]]
TYPE__flat_argument_response = Optional[Union[List[SwaggerArgumentResponse], SwaggerArgumentResponse]]
TYPE__listed_argument_response = Union[List[SwaggerArgumentResponse], SwaggerList, SwaggerSimpleList]


def validate_argument_responses(arg_responses: TYPE__argument_response):
    if isinstance(arg_responses, SwaggerList):
        arg_responses = arg_responses.responses

    if arg_responses is not None and isinstance(arg_responses, list):
        found_names = []
        for arg_resp in arg_responses:
            if isinstance(arg_resp, SwaggerList):
                validate_argument_responses(arg_resp.responses)
            else:
                try:
                    is_list = isinstance(arg_resp.arg_type, SwaggerList)
                    is_resp = isinstance(arg_resp.arg_type, SwaggerArgumentResponse)
                    is_resp_list = isinstance(arg_resp.arg_type, List)
                    if is_list or is_resp or is_resp_list:
                        validate_argument_responses(arg_resp.arg_type)

                    if arg_resp.name in found_names:
                        raise SwaggerException(ERR__duplicated_argument_response_name % arg_resp.name)
                    found_names.append(arg_resp.name)
                except:
                    pass


class SwaggerFlatResponse:

    def __init__(self, description: str = "HTTP OK", code: Union[HTTPStatus, int] = HTTPStatus.OK, resp_type: type = str,
                 body: str = RESPONSE__OK):
        self.code = code
        self.description = description
        self.body = body
        self.resp_type = resp_type

        if not isinstance(body, resp_type):
            raise SwaggerException(ERR__response_type % (code, str(resp_type)))


MOCK__description = ""


class SwaggerResponse:

    def __init__(self, description: str, code: HTTPStatus = HTTPStatus.OK, response: TYPE__argument_response = None,
                 error_on_unexpected_field: bool = False):
        self.code = code
        self.description = description
        self.error_on_unexpected_field = error_on_unexpected_field

        validate_argument_responses(response)

        self.responses = response if isinstance_union(response, TYPE__swagger_list_like) else force_list(response)


RES__allow_all = SwaggerResponse(
    description="Allows all",
    response=ARG_RESP__allow_all
)

TYPE__response = Union[SwaggerFlatResponse, SwaggerResponse]
TYPE__responses = Optional[Union[TYPE__response, List[TYPE__response]]]


class SwaggerMethod:

    def __init__(self, name: str, description: str, method: str, arguments: TYPE__flat_argument_response = None,
                 body: TYPE__argument_response = None, response: TYPE__responses = None,
                 exceptions: TYPE__exceptions = None, parallel_verification: bool = False):
        self._validate_exceptions(exceptions)
        self._validate_responses(response)
        validate_argument_responses(arguments)
        validate_argument_responses(body)

        self.name = name
        self.description = description
        self.method = method
        self.parallel_verification = parallel_verification

        self.responses = response if isinstance_union(response, TYPE__swagger_list_like) else force_list(response)
        self.arguments = force_list(arguments)
        self._validate_arguments(arguments)
        self.body = force_list(body)
        self.exceptions = force_list(exceptions)

        found_500 = any([_http_status_to_integer(ex.code) == RESP__default_err_code for ex in self.exceptions])
        if not found_500:
            self.exceptions.append(SwaggerResponseException())

        if self.method == REST__DELETE and len(self.body) != 0:
            raise Exception(ERR__delete_not_allowed_body)

        if self.method == REST__GET and len(self.body) != 0:
            raise Exception(ERR__get_not_allowed_body)

    @staticmethod
    def _validate_exceptions(exceptions: TYPE__exceptions):
        if exceptions is not None and isinstance(exceptions, list):
            found_codes = []
            for ex in exceptions:
                if _http_status_to_integer(ex.code) in found_codes:
                    raise SwaggerException(ERR__duplicated_response_exception_code % ex.code)
                found_codes.append(_http_status_to_integer(ex.code))

    @staticmethod
    def _validate_responses(responses: TYPE__responses):
        if responses is not None and isinstance(responses, list):
            found_codes = []
            for ex in responses:
                if _http_status_to_integer(ex.code) in found_codes:
                    raise SwaggerException(ERR__duplicated_response_exception_code % _http_status_to_integer(ex.code))
                found_codes.append(_http_status_to_integer(ex.code))

    @staticmethod
    def _validate_arguments(arguments: List[Union[SwaggerArgumentResponse, SwaggerList]]):
        arguments = force_list(arguments)
        for arg in arguments:
            if isinstance(arg.arg_type, SwaggerArgumentResponse):
                raise Exception(ERR__nested_not_allowed)
            elif isinstance(arg, SwaggerList):
                raise Exception(ERR__list_not_allowed)


TYPE__swagger_method = Union[List[SwaggerMethod], SwaggerMethod]


class SwaggerDocumentation:

    def __init__(self, methods: TYPE__swagger_method, tags: str = None, security: bool = True):
        self._validate_methods(methods)
        self.methods = force_list(methods)
        self.security = security
        self.path: str = PATH__empty  # Attached later to the method
        self.tags = tags

    @staticmethod
    def _validate_methods(methods: TYPE__swagger_method):
        if isinstance(methods, list):
            found_methods = []
            for method in methods:
                if method.method in found_methods:
                    raise SwaggerException(ERR__duplicated_method % method.method)
                found_methods.append(method)


def _gen_yaml_indent(indent: int) -> str:
    return ''.join(indent * [OPEN_API__spacer])


def _build_yaml(yaml: str, indent: int, attr: str, value: Union[str, List[str]] = "") -> str:
    line = _gen_yaml_indent(indent)
    if value != "" and not isinstance(value, list):
        value = " " + str(value)
    line = line + attr + YAML__separator

    if not isinstance(value, list):
        line = line + str(value) + YAML__new_line
    else:
        line += YAML__new_line
        list_indent = ''.join((indent + 1) * [OPEN_API__spacer])
        for itm in value:
            line = line + list_indent + YAML__list_separator + itm + YAML__new_line

    return yaml + line


def _http_status_to_integer(code: HTTPStatus) -> int:
    if isinstance(code, int):
        return code
    else:
        return code.value


def _insert_yaml_line(yaml: str) -> str:
    return yaml + YAML__new_line


def _to_openapi_type(python_type: type) -> str:
    if isinstance(python_type, list):
        return OPEN_API__object
    elif isinstance(python_type, SwaggerList):
        return OPEN_API__array
    elif isinstance(python_type, SwaggerArgumentResponse):
        return OPEN_API__object
    else:
        return OPEN_API__typemap[python_type]  # Lint issue PYC


def _generate_examples(yaml: str, depth: int, response: TYPE__listed_argument_response, is_prod: bool) -> str:
    if isinstance(response, SwaggerList):
        yaml = yaml + _gen_yaml_indent(depth) + YAML__list_separator + YAML__new_line
        response = response.responses
        depth += 1
    elif isinstance(response, SwaggerSimpleList):
        yaml = _build_yaml(yaml, depth, OPEN_API__example, response.example)
        return yaml

    for response_property in response:
        if isinstance(response_property, SwaggerList):
            yaml = _generate_examples(yaml, depth + 1, response_property, is_prod)
        else:
            is_complex = isinstance(response_property.arg_type, List)

            if isinstance(response_property, SwaggerArgumentResponse) and response_property.local_only and is_prod:
                continue

            if is_complex or isinstance(response_property.arg_type, SwaggerList):
                yaml = _build_yaml(yaml, depth, response_property.name)
                yaml = _generate_examples(yaml, depth + 1, response_property.arg_type, is_prod)
            elif len(response_property.example) != 0:
                raw_example = response_property.example[0]
                if response_property.arg_type == str:
                    raw_example = YAML__double_quote + str(raw_example) + YAML__double_quote
                yaml = _build_yaml(yaml, depth, response_property.name, raw_example)
            elif response_property.arg_type == ARG_RESP__allow_all:
                yaml = _build_yaml(yaml, depth, response_property.name, "{}")

    return yaml


def _generate_properties(yaml: str, depth: int, response: TYPE__listed_argument_response, is_prod: bool) -> str:
    if isinstance(response, SwaggerList):
        yaml = _build_yaml(yaml, depth, OPEN_API__items)
        response = response.responses
        depth += 1
    elif isinstance(response, SwaggerSimpleList):
        yaml = _build_yaml(yaml, depth, OPEN_API__items)
        yaml = _build_yaml(yaml, depth + 1, OPEN_API__type, _to_openapi_type(response.arg_type))
        return yaml

    required_props = [YAML__double_quote + response_property.name + YAML__double_quote
                      for response_property in response
                      if not isinstance(response_property, SwaggerList) and response_property.required]
    required = YAML__array_separator.join(required_props)

    if len(response) == 0 or not isinstance(response[0], SwaggerList):
        if len(required_props) != 0:
            yaml = _build_yaml(yaml, depth, OPEN_API__required, YAML__array_open + required + YAML__array_close)
        yaml = _build_yaml(yaml, depth, OPEN_API__properties)

    for response_property in response:
        if isinstance(response_property, SwaggerList):
            yaml = _generate_properties(yaml, depth, response_property, is_prod)
        else:
            if isinstance(response_property, SwaggerArgumentResponse) and response_property.local_only and is_prod:
                continue
            yaml = _build_yaml(yaml, depth + 1, response_property.name)
            yaml = _build_yaml(yaml, depth + 2, OPEN_API__description, response_property.description)

            is_complex = isinstance(response_property.arg_type, List) or isinstance(response_property.arg_type,
                                                                                    SwaggerList)

            yaml = _build_yaml(yaml, depth + 2, OPEN_API__type, _to_openapi_type(response_property.arg_type))

            if is_complex:
                yaml = _generate_properties(yaml, depth + 2, response_property.arg_type, is_prod)

    return yaml


def _build_parameter(yaml: str, depth: int, doc: SwaggerDocumentation, arg: SwaggerArgumentResponse,
                     is_prod: bool) -> str:
    if is_prod and arg.local_only:
        return yaml

    path_var = OPEN_API__path_var_open + arg.name + OPEN_API__path_var_close
    is_path = path_var in doc.path

    yaml = _build_yaml(yaml, depth, YAML__list_separator + OPEN_API__in,
                       OPEN_API__in_path if is_path else OPEN_API__in_query)
    yaml = _build_yaml(yaml, depth + 1, OPEN_API__name, arg.name)
    yaml = _build_yaml(yaml, depth + 1, OPEN_API__required, str(arg.required))
    if len(arg.example) != 0:
        yaml = _build_yaml(yaml, depth + 1, OPEN_API__example, str(arg.example[0]))
    yaml = _build_yaml(yaml, depth + 1, OPEN_API__description, arg.description)
    yaml = _build_yaml(yaml, depth + 1, OPEN_API__schema)
    yaml = _build_yaml(yaml, depth + 2, OPEN_API__type, _to_openapi_type(arg.arg_type))

    return yaml


def _produce_path(doc: SwaggerDocumentation, yaml: str, is_prod: bool) -> str:
    if doc.path == PATH__empty:
        raise Exception(ERR__empty_path)

    yaml = _build_yaml(yaml, 1, doc.path)

    for method in doc.methods:
        yaml = _build_yaml(yaml, 2, method.method.lower())
        if doc.tags is not None:
            yaml = _build_yaml(yaml, 3, OPEN_API__tags, [doc.tags])
        yaml = _build_yaml(yaml, 3, OPEN_API__summary, method.name)
        if doc.security:
            yaml = _build_yaml(yaml, 3, OPEN_API__parameters)
            yaml = _build_yaml(yaml, 4, YAML__list_separator + OPEN_API__in, OPEN_API__in_header)
            yaml = _build_yaml(yaml, 5, OPEN_API__name, HEADER__security)
            yaml = _build_yaml(yaml, 5, OPEN_API__example, str(EXAMPLE__jwt))
            yaml = _build_yaml(yaml, 5, OPEN_API__description, DESCRIPTION__security)
            yaml = _build_yaml(yaml, 5, OPEN_API__required, str(True))
            yaml = _build_yaml(yaml, 5, OPEN_API__schema)
            yaml = _build_yaml(yaml, 6, OPEN_API__type, TYPE__string)
        if len(method.arguments) != 0:
            if not doc.security:
                yaml = _build_yaml(yaml, 3, OPEN_API__parameters)
            for arg in method.arguments:
                yaml = _build_parameter(yaml, 4, doc, arg, is_prod)
        if isinstance(method.body, SwaggerList) or len(method.body) != 0:
            yaml = _build_yaml(yaml, 3, OPEN_API__request_body)
            yaml = _build_yaml(yaml, 4, OPEN_API__required, str(True))
            yaml = _build_yaml(yaml, 4, OPEN_API__content)
            yaml = _build_yaml(yaml, 5, CONTENT_TYPE__json)
            yaml = _build_yaml(yaml, 7, OPEN_API__schema)
            use_body = method.body
            use_depth = 8
            if isinstance(method.body, SwaggerList):
                yaml = _build_yaml(yaml, 8, OPEN_API__type, OPEN_API__array)
                yaml = _build_yaml(yaml, 8, OPEN_API__example)
                yaml = _generate_examples(yaml, 9, method.body, is_prod)
                yaml = _build_yaml(yaml, 8, OPEN_API__items)
                yaml = _build_yaml(yaml, 9, OPEN_API__type, OPEN_API__object)
                use_body = method.body.responses
                use_depth = 9
            else:
                yaml = _build_yaml(yaml, 8, OPEN_API__type, OPEN_API__object)
                yaml = _build_yaml(yaml, 8, OPEN_API__example)
                yaml = _generate_examples(yaml, 9, method.body, is_prod)
            yaml = _generate_properties(yaml, use_depth, use_body, is_prod)
        yaml = _build_yaml(yaml, 3, OPEN_API__description, method.description)
        yaml = _build_yaml(yaml, 3, OPEN_API__responses)

        if not any([resp.code == HTTPStatus.OK for resp in method.responses]):
            method.responses.append(SwaggerFlatResponse(RESPONSE__200_ok))

        for response in method.responses:
            yaml = _build_yaml(yaml, 4, YAML__quote + str(_http_status_to_integer(response.code)) + YAML__quote)
            yaml = _build_yaml(yaml, 5, OPEN_API__description, response.description)
            yaml = _build_yaml(yaml, 5, OPEN_API__content)
            yaml = _build_yaml(yaml, 6, CONTENT_TYPE__json)
            yaml = _build_yaml(yaml, 7, OPEN_API__schema)
            if isinstance(response, SwaggerFlatResponse):
                yaml = _build_yaml(yaml, 8, OPEN_API__type, _to_openapi_type(response.resp_type))
            if isinstance(response, SwaggerResponse):
                use_response = response.responses
                use_depth = 8
                if isinstance(response.responses, SwaggerList):
                    yaml = _build_yaml(yaml, 8, OPEN_API__type, OPEN_API__array)
                    yaml = _build_yaml(yaml, 8, OPEN_API__example)
                    yaml = _generate_examples(yaml, 9, response.responses, is_prod)
                    yaml = _build_yaml(yaml, 8, OPEN_API__items)
                    if isinstance(response.responses.responses[0], SwaggerList):
                        yaml = _build_yaml(yaml, 9, OPEN_API__type, OPEN_API__array)
                    else:
                        yaml = _build_yaml(yaml, 9, OPEN_API__type, OPEN_API__object)
                    use_depth = 9
                    use_response = response.responses.responses
                elif isinstance(response.responses, SwaggerSimpleList):
                    yaml = _build_yaml(yaml, 8, OPEN_API__type, OPEN_API__array)
                    yaml = _generate_examples(yaml, 8, response.responses, is_prod)
                else:
                    yaml = _build_yaml(yaml, 8, OPEN_API__type, OPEN_API__object)
                    yaml = _build_yaml(yaml, 8, OPEN_API__example)
                    yaml = _generate_examples(yaml, 9, response.responses, is_prod)
                yaml = _generate_properties(yaml, use_depth, use_response, is_prod)
            else:
                yaml = _build_yaml(yaml, 8, OPEN_API__example, str(response.body))

        for exception in method.exceptions:
            yaml = _build_yaml(yaml, 4, YAML__quote + str(_http_status_to_integer(exception.code)) + YAML__quote)
            yaml = _build_yaml(yaml, 5, OPEN_API__description, exception.message)

    return yaml


def _produce_documentation(docs: ModuleType, url: str, base_path: str, is_prod: bool = False):
    title: str
    description: str
    version: str
    filename: str

    try:
        output = bool(getattr(docs, ATTR__output))
        if not output:
            return
    except AttributeError:
        pass  # Assume that output is True when cannot find flag

    try:
        title = str(getattr(docs, ATTR__title))
    except AttributeError:
        raise Exception(ERR__documentation_variable_expected % (ATTR__title, docs.__name__))
    try:
        version = str(getattr(docs, ATTR__version))
    except AttributeError:
        version = VERSION
    try:
        description = str(getattr(docs, ATTR__description))
    except AttributeError:
        raise Exception(ERR__documentation_variable_expected % (ATTR__description, docs.__name__))
    try:
        filename = str(getattr(docs, ATTR__filename))
    except AttributeError:
        raise Exception(ERR__documentation_variable_expected % (ATTR__filename, docs.__name__))

    yaml = _build_yaml("", 0, OPEN_API__openapi, VERSION__open_api)
    yaml = _build_yaml(yaml, 0, OPEN_API__info)
    yaml = _build_yaml(yaml, 1, OPEN_API__title, title)
    yaml = _build_yaml(yaml, 1, OPEN_API__description, description)
    yaml = _build_yaml(yaml, 1, OPEN_API__version, version)
    yaml = _insert_yaml_line(yaml)

    yaml = _build_yaml(yaml, 0, OPEN_API__servers)
    yaml = _build_yaml(yaml, 1, YAML__list_separator + OPEN_API__url, url)
    yaml = _insert_yaml_line(yaml)

    yaml = _build_yaml(yaml, 0, OPEN_API__paths)

    all_docs = [getattr(docs, item) for item in docs.__dict__ if isinstance(getattr(docs, item), SwaggerDocumentation)]

    for documentation in all_docs:
        yaml = _produce_path(documentation, yaml, is_prod)

    swagger_output_file = open(os.path.join(base_path, filename + EXTENSION__yaml), "w")
    swagger_output_file.write(yaml)
    swagger_output_file.close()

    yaml = yaml_utils.safe_load(yaml)

    swagger_template_file = open(os.path.join(get_jaaql_root(), DIR__scripts, FILE__swagger_template), "r")
    swagger_template = swagger_template_file.read()
    swagger_template_file.close()

    template_replace = TEMPLATE__url % (filename + EXTENSION__json) if is_prod else TEMPLATE__spec + json.dumps(yaml)
    swagger_template = swagger_template % (title, template_replace)
    swagger_output_file = open(os.path.join(base_path, filename + EXTENSION__html), "w")
    swagger_output_file.write(swagger_template)
    swagger_output_file.close()
    swagger_json_output_file = open(os.path.join(base_path, filename + EXTENSION__json), "w")
    swagger_json_output_file.write(json.dumps(yaml))
    swagger_json_output_file.close()


def produce_all_documentation(all_docs: List[ModuleType], url: str, is_prod: bool = False, base_path: str = None):
    create_dir = DIR__swagger
    if base_path is not None:
        create_dir = os.path.join(base_path, create_dir)

    if os.path.exists(create_dir):
        shutil.rmtree(create_dir)
    os.makedirs(create_dir)

    for docs in all_docs:
        _produce_documentation(docs, url, create_dir, is_prod)



from jaaql.exceptions.http_status_exception import *
from datetime import datetime
import re
from jaaql.utilities.crypt_utils import encrypt_raw, decrypt_raw
import uuid
import queue
from jaaql.db.db_interface import DBInterface, ECHO__none
from jaaql.constants import KEY__position, KEY__file, KEY__application, KEY__configuration, KEY__error, KEY__error_row_number, KEY__error_query, \
    KEY__error_set, KEY__error_index
from typing import Union


ERR_malformed_statement = "Malformed query, expecting string or dictionary"
ERR_unknown_assert = "Unknown assert type '%s'. Please use one of %s"
ERR_malformed_parameters = "Malformed parameters! Please send as dictionary"
ERR_malformed_decrypt = "Malformed decryption request! Please send as a list of columns"
ERR_missing_parameter = "Missing parameter! Expected to find '%s'"
ERR_missing_query = "Missing query key from input dictionary"
ERR_unused_parameter = "Unused parameter! Supplied with '%s' but was never used"
ERR_polluted_input = "Input polluted. If passing query, only pass parameters as well"
ERR_malformed_query = "Value of 'query' key malformed. Please use a string if you are passing a value to 'query'"
ERR_mistyped_parameter = "Type '%s', for parameter '%s' unexpected. Please provide either a string, float or integer"
ERR_malformed_operation_type = "Operation malformed. Expecting either a list, string or dictionary input"
ERR_assert_expecting = "Assert expecting %s row(s) but received %d row(s)!"
ERR_malformed_join = "Joins only allowed as list or string input at the moment!"
ERR__missing_decrypt_column = "Decrypted column '%s' not found in the result set"
ERR_missing_store = "Missing parameter for store operation '%s'"
ERR_store_parameter_malformatted = "Parameter '%s' for store operation should be a list"

KEY_parameters = "parameters"
KEY_decrypt = "decrypt"
KEY_select = "select"
KEY_where = "where"
KEY_echo = "echo"
KEY_query = "query"
KEY_autocommit = "autocommit"
KEY_order_by = "order by"
KEY_group_by = "group by"
KEY_assert = "assert"
KEY_join = "join"
KEY_join_left = "left join"
KEY_join_cross = "cross join"
KEY_join_inner = "inner join"
KEY_join_right = "right join"
KEY_join_full = "full join"
KEY_join_lateral = "lateral join"
KEY_store = "store"
KEY__state = "_state"
KEY_state = "state"
KEY__row_number = "_rowNumber"

STATEMENT_as = "as"
STATEMENT_argument_begin = "%"
STATEMENT_paren_open = "("
STATEMENT_paren_close = ")"
STATEMENT_space = " "
STATEMENT_empty = ""
STATEMENT_comma = ","
STATEMENT_new_line = "\r\n"
STATEMENT_null = "NULL"

ASSERT_none = None
ASSERT_zero = 0
ASSERT_one = 1
ASSERT_one_plus = "1+"
ASSERT_one_plus_minimum_allowed = 2
ASSERT_one_plus_message = "more than 1"
ASSERT_allowed = [ASSERT_zero, ASSERT_one, ASSERT_one_plus]

REGEX_query_argument = r':([a-zA-Z0-9_.\-])+(?=[^\']*(?:\'[^\']*\'[^\']*)*$)'  # Match all :blah not in quotes
REGEX_enc_query_argument = r'#([a-zA-Z0-9_.\-])+(?=[^\']*(?:\'[^\']*\'[^\']*)*$)'  # Match all #blah not in quotes
REGEX_enc_literals = r'#\'(([^\']*)(\'\')*)+\''

PYFORMAT_str = "s"

OPERATOR_none = None
OPERATOR_separators = {
    KEY_select: STATEMENT_comma,
    KEY_where: STATEMENT_space,
    KEY_order_by: STATEMENT_comma,
    KEY_group_by: STATEMENT_comma
}
OPERATOR_joins = [
    KEY_join,
    KEY_join_left,
    KEY_join_cross,
    KEY_join_inner,
    KEY_join_right,
    KEY_join_full,
    KEY_join_lateral
]
SEPARATOR_none = ""
SEPARATOR_space = " "
SEPARATOR_spaces_per_tab = 4

MARKER_DEFAULT = "default"
MARKERS = [MARKER_DEFAULT]


class InterpretJAAQL:

    def __init__(self, db_interface: DBInterface):
        self.db_interface = db_interface

    def transform(self, operation: Union[dict, str], conn=None, skip_commit: bool = False, wait_hook: queue.Queue = None,
                  encryption_key: bytes = None, autocommit: bool = False, canned_query_service=None):
        if (not isinstance(operation, dict)) and (not isinstance(operation, str)):
            raise HttpStatusException(ERR_malformed_operation_type, HTTPStatus.BAD_REQUEST)

        was_conn_none = False
        if conn is None:
            conn = self.db_interface.get_conn()
            was_conn_none = True

        is_dict_operation = isinstance(operation, dict)
        query = {"query": {KEY_query: operation, KEY_assert: None, KEY_decrypt: None, KEY_parameters: {}}}
        is_dict_query = False
        parameters = {}
        past_parameters = {}
        check_required = True
        ret = {}
        err = None

        if autocommit:
            conn.autocommit = autocommit

        if is_dict_operation:
            query = operation.get(KEY_query)
            if KEY_autocommit in operation:
                if conn.autocommit != operation[KEY_autocommit]:
                    conn.commit()
                conn.autocommit = operation[KEY_autocommit]

            if query is None:
                canned_query = canned_query_service.get_canned_query(operation[KEY__application], operation[KEY__configuration],
                                                                     operation[KEY__file], operation[KEY__position])
                query = {"query": {KEY_query: canned_query, KEY_assert: operation.get(KEY_assert), KEY_decrypt: operation.get(KEY_decrypt),
                                   KEY_parameters: {}}}
                check_required = False
            else:
                is_dict_query = isinstance(query, dict)
                if not is_dict_query:
                    query = {"query": {"query": query, KEY_assert: operation.get(KEY_assert), KEY_decrypt: operation.get(KEY_decrypt),
                                       KEY_parameters: {}}}
                else:
                    all_replaced = True
                    for key, val in query.items():
                        if isinstance(val, str):
                            all_replaced = False
                            query[key] = {KEY_query: val, KEY_assert: None, KEY_decrypt: None, KEY_parameters: {}}
                        else:
                            if KEY_store in val:
                                for sub_key, sub_val in val.items():
                                    if sub_key in [KEY_parameters, KEY_decrypt, KEY_store]:
                                        continue
                                    if isinstance(sub_val, str):
                                        all_replaced = False
                                    else:
                                        canned_query = canned_query_service.get_canned_query(operation[KEY__application],
                                                                                             operation[KEY__configuration],
                                                                                             sub_val[KEY__file], sub_val[KEY__position])
                                        val[sub_key] = canned_query
                                if KEY_parameters not in val:
                                    val[KEY_parameters] = {}
                                if KEY_decrypt not in val:
                                    val[KEY_decrypt] = None
                            else:
                                fetched_query = None
                                if isinstance(val[KEY_query], str):
                                    all_replaced = False
                                    fetched_query = val[KEY_query]
                                else:
                                    fetched_query = canned_query_service.get_canned_query(operation[KEY__application],
                                                                                          operation[KEY__configuration],
                                                                                          val[KEY_query][KEY__file], val[KEY_query][KEY__position])
                                query[key] = {KEY_query: fetched_query, KEY_assert: val.get(KEY_assert), KEY_decrypt: val.get(KEY_decrypt),
                                              KEY_parameters: val.get(KEY_parameters, {})}
                    check_required = not all_replaced

            parameters = operation.get(KEY_parameters, {})

        unused_orig_parameters = set(parameters.keys())

        was_store = None
        exc_query_key = None
        exc_row_idx = None
        exc_parameters = None
        exc_query = None
        exc_row_number = None
        exc_state = None

        try:
            for query_key, query_obj in query.items():
                exc_query_key = query_key
                replacement_parameters = {**past_parameters, **parameters, **query_obj[KEY_parameters]}

                to_exec = []
                states = []
                was_store = is_dict_query and KEY_store in query_obj
                if was_store:
                    if query_obj[KEY_store] not in replacement_parameters:
                        raise HttpStatusException(ERR_missing_store % query_obj[KEY_store])
                    if not isinstance(replacement_parameters[query_obj[KEY_store]], list):
                        raise HttpStatusException(ERR_store_parameter_malformatted % query_obj[KEY_store])
                    unused_orig_parameters -= {query_obj[KEY_store]}
                    to_exec = [query_obj[obj[KEY__state]] for obj in replacement_parameters[query_obj[KEY_store]]]
                    states = [obj[KEY__state] for obj in replacement_parameters[query_obj[KEY_store]]]
                    iterate_params = replacement_parameters.pop(query_obj[KEY_store])
                    replacement_parameters = [{**replacement_parameters, **{query_obj[KEY_store] + "." + key: val for key, val in cur.items()}}
                                              for cur in iterate_params]
                else:
                    to_exec = [query_obj[KEY_query]]
                    replacement_parameters = [replacement_parameters]
                    states = [None] * len(to_exec)

                if was_store:
                    ret[query_key] = []

                for cur_query, cur_parameters, cur_row_idx, cur_state in zip(to_exec, replacement_parameters, range(len(to_exec)), states):
                    exc_query = cur_query
                    exc_state = cur_state
                    exc_row_idx = cur_row_idx
                    exc_parameters = cur_parameters
                    exc_row_number = cur_parameters.get(KEY__row_number)
                    last_query, found_parameter_dictionary = self.pre_prepare_statement(cur_query, cur_parameters)

                    enc_parameter_dictionary = {}

                    encrypt_parameters = {key: val for key, val in cur_parameters.items() if key not in found_parameter_dictionary}

                    if len(encrypt_parameters) != 0:
                        last_query, enc_parameter_dictionary = self.pre_prepare_statement(last_query, encrypt_parameters,
                                                                                          match_regex=REGEX_enc_query_argument,
                                                                                          encryption_key=encryption_key)

                    last_query = self.encrypt_literals(last_query, encryption_key)
                    found_params = {**found_parameter_dictionary, **enc_parameter_dictionary}
                    unused_orig_parameters -= set(found_params.keys())

                    res = self.db_interface.execute_query_fetching_results(conn, last_query, found_params, wait_hook=wait_hook,
                                                                           requires_dba_check=check_required and canned_query_service is not None)
                    wait_hook = None  # We've already checked authentication, we don't need to do it again
                    check_required = False  # We've done the first check, if we have reached here without an exception, user does not need a DBA check

                    if len(res['rows']) == 1 and not was_store:
                        for column, row in zip(res['columns'], res['rows'][0]):
                            past_parameters[query_key + "." + column] = row

                    if is_dict_query:
                        if was_store:
                            ret[query_key].append(res)
                        else:
                            ret[query_key] = res
                    else:
                        ret = res

                    cur_assert = query_obj.get(KEY_assert)
                    if cur_assert != ASSERT_none:
                        if cur_assert == ASSERT_zero and len(res["rows"]) != ASSERT_zero:
                            raise HttpStatusException(ERR_assert_expecting % (str(ASSERT_zero), len(res["rows"])), HTTPStatus.BAD_REQUEST)
                        elif cur_assert == ASSERT_one and len(res["rows"]) != ASSERT_one:
                            raise HttpStatusException(ERR_assert_expecting % (str(ASSERT_one), len(res["rows"])), HTTPStatus.BAD_REQUEST)
                        elif cur_assert == ASSERT_one_plus and len(res["rows"]) < ASSERT_one_plus_minimum_allowed:
                            raise HttpStatusException(ERR_assert_expecting % (str(ASSERT_one_plus_message), len(res["rows"])), HTTPStatus.BAD_REQUEST)

                    if query_obj[KEY_decrypt] is not None:
                        missing = [param for param in query_obj[KEY_decrypt] if param not in res["columns"]]
                        if len(missing) != 0:
                            raise HttpStatusException(ERR__missing_decrypt_column % missing)

                        if len(query_obj[KEY_decrypt]) != 0:
                            res["rows"] = [
                                [decrypt_raw(encryption_key, val) if col in query_obj[KEY_decrypt] and val is not None else val for val, col in
                                 zip(row, res["columns"])] for row in res["rows"]]

            if len(unused_orig_parameters):
                raise HttpStatusException(ERR_unused_parameter % list(unused_orig_parameters)[0], HTTPStatus.BAD_REQUEST)
        except Exception as ex:
            if is_dict_query:
                new_ex = HttpStatusException(str(ex))
                if isinstance(ex, HttpStatusException):
                    new_ex.response_code = new_ex.response_code
                ex = new_ex

                ex.message = {
                    KEY__error: ex.message,
                    KEY__error_set: exc_query_key,
                    KEY__error_query: exc_query,
                    KEY_parameters: exc_parameters
                }
                if was_store:
                    ex.message[KEY__error_row_number] = exc_row_number
                    ex.message[KEY__error_index] = exc_row_idx
                    ex.message[KEY_state] = exc_state

            err = ex

        if was_conn_none:
            self.db_interface.put_conn_handle_error(conn, err, STATEMENT_empty, skip_rollback_commit=skip_commit)
        else:
            self.db_interface.handle_error(conn, err, STATEMENT_empty)

        return ret

    def render_cur_operation(self, operation):
        if isinstance(operation, dict):
            echo = ECHO__none
            assert_ = ASSERT_none
            parameters = {}
            decrypt_columns = {}

            if KEY_query not in operation:
                raise HttpStatusException(ERR_missing_query, HTTPStatus.BAD_REQUEST)

            if KEY_parameters in operation:
                if not isinstance(operation[KEY_parameters], dict):
                    raise HttpStatusException(ERR_malformed_parameters, HTTPStatus.BAD_REQUEST)
                parameters = operation[KEY_parameters]

            if KEY_decrypt in operation:
                if not isinstance(operation[KEY_decrypt], list):
                    raise HttpStatusException(ERR_malformed_parameters, HTTPStatus.BAD_REQUEST)
                decrypt_columns = operation[KEY_decrypt]

            if KEY_assert in operation:
                assert_ = operation[KEY_assert]
                if KEY_assert not in ASSERT_allowed:
                    raise HttpStatusException(ERR_unknown_assert % (assert_, ", ".join(ASSERT_allowed)),
                                              HTTPStatus.BAD_REQUEST)

            if KEY_echo in operation:
                echo = operation[KEY_echo]

            query = self.render_dict_operation(operation[KEY_query])
            return query, parameters, decrypt_columns, echo, assert_
        elif isinstance(operation, str):
            return operation, {}, ECHO__none, ASSERT_none
        else:
            raise HttpStatusException(ERR_malformed_statement, HTTPStatus.BAD_REQUEST)

    def render_dict_operation(self, query):
        if isinstance(query, dict):
            return self.render_statement_from_dict(query)
        elif isinstance(query, str):
            return query
        else:
            raise HttpStatusException(ERR_malformed_statement, HTTPStatus.BAD_REQUEST)

    def get_format_type(self, key, value):
        # There is not an error below. psycopg2 _requires_ that the format is always string
        if isinstance(value, str):
            return PYFORMAT_str
        elif isinstance(value, bytes):
            return PYFORMAT_str
        elif isinstance(value, int):
            return PYFORMAT_str
        elif isinstance(value, float):
            return PYFORMAT_str
        elif isinstance(value, datetime):
            return PYFORMAT_str
        elif isinstance(value, uuid.UUID):
            return PYFORMAT_str
        else:
            raise HttpStatusException(ERR_mistyped_parameter % (type(value).__name__, key), HTTPStatus.BAD_REQUEST)

    def render_statement(self, statement):
        parameters = {}

        if isinstance(statement, str):
            query = statement
        elif isinstance(statement, dict):
            query, parameters = self.render_statement_from_dict(statement)
        else:
            raise HttpStatusException(ERR_malformed_statement, HTTPStatus.BAD_REQUEST)

        return query, parameters

    def render_statement_from_dict(self, statement, depth=0):
        query = ""

        custom_operators = {
            KEY_select: self.render_select
        }

        i = 0
        for key, val in statement.items():
            trans_key = key.lower()

            if i != 0:
                query += STATEMENT_new_line

            query += trans_key.upper()

            render_function = custom_operators.get(trans_key, self.render_element)
            if trans_key in OPERATOR_joins:
                render_function = self.render_join

            unjoined = render_function(val, trans_key, depth)
            query += self.join_line_with_separator(trans_key, 1, unjoined, trans_key in OPERATOR_joins)

            i += 1

        return query

    def get_spaces_depth(self, depth):
        spaces = [SEPARATOR_space] * depth * SEPARATOR_spaces_per_tab
        return SEPARATOR_none.join(spaces)

    def join_line_with_separator(self, operator, depth, elements, is_join=False):
        spaces = self.get_spaces_depth(depth)

        joiner = OPERATOR_separators.get(operator, SEPARATOR_none) + STATEMENT_new_line
        if not is_join:
            joiner += spaces
        joined = joiner.join([str(element) for element in elements])

        return STATEMENT_new_line + spaces + joined

    def render_join(self, element, operator, depth=0, do_join=False):
        spaces = self.get_spaces_depth(depth + 1)

        if isinstance(element, str):
            return element
        elif isinstance(element, list):
            return [cur if ix == 0 else operator.upper() + STATEMENT_new_line + spaces + cur for ix, cur in zip(
                range(len(element)), element)]
        else:
            raise HttpStatusException(ERR_malformed_join, HTTPStatus.BAD_REQUEST)

    def render_select(self, element, operator, depth=0, do_join=False):
        if isinstance(element, dict):
            return [
                self.join_line_with_separator(KEY_select, depth, self.render_element(val, OPERATOR_none, depth))
                + " " + STATEMENT_as + " \"" + key + "\"" for key, val in element.items()]
        else:
            return self.render_element(element, KEY_select, depth)

    def render_element(self, element, operator, depth=0, do_join=False):
        if isinstance(element, list):
            unjoined = [
                self.render_element(sub_element, operator, depth + 1)
                if isinstance(sub_element, list) or isinstance(sub_element, dict) else sub_element
                for sub_element in element
            ]
            return unjoined
        elif isinstance(element, dict):
            opener = STATEMENT_paren_open if depth != 0 else ""
            closer = STATEMENT_paren_close if depth != 0 else ""
            unjoined = [
                opener + self.render_element(operands, " " + new_operator + " ", depth + 1) + closer
                for new_operator, operands in element.items()
            ]
            return unjoined
        else:
            return [element]

    def encrypt_literals(self, query, encryption_key: bytes = None, match_regex: str = REGEX_enc_literals):
        new_query = ""
        last_start_match = 0
        last_end_match = 0

        for match in re.finditer(match_regex, query):
            start_pos = match.regs[0][0]
            end_pos = match.regs[0][1]
            match_str = query[start_pos + 2:end_pos - 1]
            new_query += query[last_start_match:start_pos] + "'" + encrypt_raw(encryption_key, match_str) + "'"
            last_start_match = start_pos
            last_end_match = end_pos

        return new_query + query[last_end_match:]

    def pre_prepare_statement(self, query, parameters, match_regex: str = REGEX_query_argument, encryption_key: bytes = None):
        prepared = ""
        last_index = 0
        found_parameters = []
        found_parameter_dictionary = {}

        for match in re.finditer(match_regex, query):
            is_skipped_default = False
            start_pos = match.regs[0][0]
            if start_pos != 0:
                if query[start_pos - 1] == REGEX_query_argument[0]:
                    continue
            end_pos = match.regs[0][1]
            match_str = query[start_pos + 1:end_pos]

            prepared += query[last_index:start_pos]
            last_index = end_pos

            if match_str not in found_parameters:
                if match_str not in parameters and match_str not in MARKERS:
                    raise HttpStatusException(ERR_missing_parameter % match_str, HTTPStatus.BAD_REQUEST)
                elif match_str not in parameters and match_str in MARKERS:
                    is_skipped_default = True
                else:
                    found_parameters.append(match_str)
                    param = parameters[match_str]
                    if encryption_key is not None:
                        param = encrypt_raw(encryption_key, param)
                    found_parameter_dictionary[match_str] = param

            if parameters[match_str] is None:
                prepared += STATEMENT_null
            else:
                type = self.get_format_type(match_str, parameters[match_str])
                if is_skipped_default:
                    prepared += MARKER_DEFAULT
                else:
                    prepared += STATEMENT_argument_begin + STATEMENT_paren_open + match_str + STATEMENT_paren_close + type

        prepared += query[last_index:]

        return prepared, found_parameter_dictionary



import os
import re
import threading

from typing import Callable
from jaaql.email.email_manager_service import Email, TYPE__email_attachments
import requests
from os.path import join
from jaaql.exceptions.http_status_exception import HttpStatusException
from jaaql.constants import *
from urllib.parse import quote
from jaaql.utilities.utils_no_project_imports import load_artifact
import time

REGEX__email_parameter = r'({{)([a-zA-Z0-9_\-]+)(}})'
REGEX__email_uri_encoded_parameter = r'(\[\[)([a-zA-Z0-9_\-]+)(\]\])'
REPLACE__str = "{{%s}}"
REPLACE__uri_encoded_str = "[[%s]]"
ERR__missing_parameter = "Missing parameter from template '%s'"
ERR__unexpected_parameter_in_template = "Unexpected parameter in template '%s'"


class EmailManager:

    def __init__(self, is_container: bool):
        self.is_container = is_container

    def send_email(self, email: Email):
        requests.post("http://127.0.0.1:" + str(PORT__ems) + ENDPOINT__send_email, json=email.repr_json())

    def wait_and_then_reload_service(self):
        time.sleep(0.5)
        requests.post("http://127.0.0.1:" + str(PORT__ems) + ENDPOINT__reload_accounts)

    def reload_service(self):
        threading.Thread(target=self.wait_and_then_reload_service).start()

    @staticmethod
    def uri_encode_replace(val):
        return quote(str(val))

    def construct_and_send_email(self, artifact_base_uri: str, template: dict, sender: str, to_email: str, to_name: str,
                                 parameters: dict = None, optional_parameters: dict = None, attachments: TYPE__email_attachments = None,
                                 attachment_access_token: str = None):
        loaded_template = load_artifact(self.is_container, artifact_base_uri, template[KEY__app_relative_path])
        if loaded_template is None:
            return

        subject, loaded_template = self.perform_replacements(template[KEY__subject], loaded_template, REPLACE__str, str, REGEX__email_parameter,
                                                             parameters, optional_parameters)
        subject, loaded_template = self.perform_replacements(subject, loaded_template, REPLACE__uri_encoded_str, EmailManager.uri_encode_replace,
                                                             REGEX__email_uri_encoded_parameter, parameters, optional_parameters)

        if attachments is not None:
            attachment_list = attachments
            if not isinstance(attachment_list, list):
                attachment_list = [attachment_list]
            for attachment in attachment_list:
                attachment.format_attachment_url(artifact_base_uri, self.is_container)

        self.send_email(Email(str(sender), template[KEY__application], str(template[KEY__email_template_name]), str(template[KEY__account]),
                              to_email, to_name, subject=subject, body=loaded_template, is_html=True, attachments=attachments,
                              attachment_access_token=attachment_access_token, override_send_name=template[KEY__override_send_name]))

    def perform_replacements(self, subject: str, html_template: str, replace_str: str, replace_func: Callable, replace_regex: str, args: dict = None,
                             optional_args: dict = None):
        if args is None:
            args = {}
        if optional_args is None:
            optional_args = {}

        for key, val in args.items():
            new_template = html_template.replace(replace_str % key.upper(), replace_func(val))
            new_subject = subject.replace(replace_str % key.upper(), replace_func(val))
            if new_template == html_template and subject == new_subject:
                raise HttpStatusException(ERR__unexpected_parameter_in_template % key)
            html_template = new_template
            subject = new_subject

        for key, val in optional_args.items():
            subject = subject.replace(replace_str % key.upper(), replace_func(val))
            html_template = html_template.replace(replace_str % key.upper(), replace_func(val))

        matched = re.findall(replace_regex, html_template)
        matched = matched + re.findall(replace_regex, subject)
        if len(matched) != 0:
            raise HttpStatusException(ERR__missing_parameter % matched[0][1])

        return subject, html_template



from jaaql.mvc.base_controller import BaseJAAQLController
from abc import abstractmethod, ABC


class JAAQLControllerInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def route(self, base_controller: BaseJAAQLController):
        pass



import os
from jaaql.utilities import crypt_utils
import json

FILE__vault = "vault"
FILE__vault_salt = "vault_salt"

DIR__vault = "vault"

ENCODING_utf = "UTF-8"


class Vault:

    def __init__(self, vault_key: str, vault_dir: str):
        crypt_utils.validate_password(vault_key)

        if not os.path.exists(vault_dir) or not os.path.isdir(vault_dir):
            os.makedirs(vault_dir)

        self._vault_dir = vault_dir
        self._vault_key = self._gen_key(vault_key)
        self._vault = self._load()

    def _load(self):
        vault_path = self._get_store_path(FILE__vault)

        vault = {}

        if os.path.exists(vault_path):
            f = open(vault_path, "rb")
            vault = json.loads(crypt_utils.decrypt(self._vault_key, f.read().decode(ENCODING_utf)))
        else:
            self._vault = {}
            self.insert_obj("init", "val")

        return vault

    def reload(self):
        self._load()

    def _persist(self):
        data = crypt_utils.encrypt(self._vault_key, json.dumps(self._vault)).encode(ENCODING_utf)
        f = open(self._get_store_path(FILE__vault), "wb")
        f.write(data)
        f.close()

    def insert_obj(self, key: str, value: str):
        self._vault[key] = value
        self._persist()
        self._load()

    def purge_object(self, key: str):
        self._vault.pop(key)
        self._persist()
        self._load()

    def has_obj(self, key: str) -> bool:
        return key in self._vault

    def get_obj(self, key: str):
        return self._vault[key]

    def _get_store_path(self, file: str) -> str:
        return os.path.join(self._vault_dir, file)

    def _gen_key(self, vault_key: str) -> bytes:
        salt = None
        vault_salt_path = self._get_store_path(FILE__vault_salt)

        if os.path.exists(vault_salt_path):
            f = open(vault_salt_path, "rb")
            salt = f.read()
            f.close()

        salt, vault_key = crypt_utils.key_stretcher(vault_key, salt)

        if not os.path.exists(vault_salt_path):
            f = open(vault_salt_path, "wb")
            f.write(salt)
            f.close()

        return vault_key



from .base_component import *


class JAAQLComponent(BaseComponent):

    def test_jaaql(self):
        pass

    # test submit, test submitting incorrect arguments incorrect types, incorrect url, rate limiting


from .base_component import *
import requests
from jaaql.constants import *


class ApplicationComponent(BaseComponent):

    def test_application_endpoints(self):
        auth_header = {HEADER_AUTH: self.get_jaaql_auth_header()}
        self.run_test_create_application(auth_header)
        self.run_test_update_application(auth_header)
        self.run_test_delete_application(auth_header)
        self.run_test_fetch_applications(auth_header)

    def run_test_create_application(self, auth_header):
        resp = requests.post(BASE_INTERNAL_URL + "/applications", json={
            "name": "CTP-App",
            "description": "The CTP App",
            "url": "apps/ctp",
            "public_username": "ctp_app"
        }, headers=auth_header)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Create app")
        resp = requests.get(BASE_URL + "/applications/public-user?application=CTP-App")
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Fetch public user")
        resp = self.get_jaaql_auth_header("ctp_app", resp.json()[KEY__password], run_test=False)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Login public user account")

        requests.post(BASE_INTERNAL_URL + "/applications", json={
            "name": "App1",
            "description": "The app called 1",
            "url": "apps/app1"
        }, headers=auth_header)
        resp = requests.get(BASE_URL + "/applications/public-user?application=App1")
        self.assertEqual(HTTPStatus.UNPROCESSABLE_ENTITY, resp.status_code, "Fail fetching non-existent public user")

        requests.post(BASE_INTERNAL_URL + "/applications", json={
            "name": "App2",
            "description": "The app called 2",
            "url": "apps/app2"
        }, headers=auth_header)

    def run_test_update_application(self, auth_header):
        resp = requests.put(BASE_INTERNAL_URL + "/applications", json={
            "name": "App1",
            "new_name": "App1Updated",
            "new_description": "App 1 but updated",
            "new_url": "{{DEFAULT}}/app1"
        }, headers=auth_header)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Update application")

    def run_test_delete_application(self, auth_header):
        resp = requests.delete(BASE_INTERNAL_URL + "/applications?name=App2", headers=auth_header)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Deletion first step")
        resp = requests.post(BASE_INTERNAL_URL + "/applications/confirm-deletion", json=resp.json(), headers=auth_header)
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Deletion second step")

    def run_test_fetch_applications(self, auth_header):
        resp = requests.get(BASE_INTERNAL_URL + "/applications?sort=name%20ASC", headers=auth_header)
        data = resp.json()[KEY__data]
        self.assertEqual(HTTPStatus.OK, resp.status_code, "Returned applications")

        self.assertEqual(5, len(data), "Records returned count")

        self.assertEqual("App1Updated", data[0][KEY__application_name], "Match second name")
        self.assertEqual("App 1 but updated", data[0][KEY__description], "Match second description")
        self.assertEqual("http://127.0.0.1/apps/app1", data[0][KEY__application_url], "Match second url")
        self.assertEqual("CTP-App", data[2][KEY__application_name], "Match first name")
        self.assertEqual("The CTP App", data[2][KEY__description], "Match first description")
        self.assertEqual("apps/ctp", data[2][KEY__application_url], "Match first url")



from jaaql.openapi.swagger_documentation import SwaggerDocumentation, SwaggerMethod, SwaggerArgumentResponse, SwaggerResponse, SwaggerList,\
    SwaggerFlatResponse, REST__POST
from jaaql.constants import *
from typing import Union, List

import copy

OUTPUT = False

ENDPOINT__refresh = "/oauth/refresh"

ARG_RES__application = SwaggerArgumentResponse(
    name=KEY__application,
    description="The application",
    example=["playground"],
    required=False,
    arg_type=str,
    condition="If this is a public user"
)

ARG_RES__configuration = SwaggerArgumentResponse(
    name=KEY__configuration,
    description="The application configuration",
    example=["main"],
    arg_type=str
)

ARG_RES__deletion_key = SwaggerArgumentResponse(
    name=KEY__deletion_key,
    description="Single use deletion key",
    arg_type=str,
    example=["eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJkZWxldGlvbl9wdXJwb3NlIjoiZXhhbXBsZSJ9.aJEPcEJBkyM9vEKT_D6AXT-hZAZ"
             "9UvagKGtljvsBj8w"],
    required=True
)

RES__deletion_key = SwaggerResponse(
    description="Returns a single use key that must be sent to the server to confirm. Key expires in 30 seconds",
    response=ARG_RES__deletion_key
)

ARG_RES__filtered_records = [
    SwaggerArgumentResponse(
        name=KEY__records_total,
        description="Total number of records without any searches etc.",
        required=True,
        arg_type=int,
        example=[100, 200]
    ),
    SwaggerArgumentResponse(
        name=KEY__records_filtered,
        description="Number of filtered records, after searches",
        required=True,
        arg_type=int,
        example=[80, 160]
    )
]


def rename_arg(arg_res: SwaggerArgumentResponse, new_name: str = None, new_description: str = None, new_examples=None):
    return SwaggerArgumentResponse(
        arg_res.name if new_name is None else new_name,
        arg_res.description if new_description is None else new_description,
        arg_res.arg_type,
        arg_res.example if new_examples is None else new_examples,
        arg_res.required,
        arg_res.condition
    )


def set_required(arg_res: SwaggerArgumentResponse, new_name: str = None, new_description: str = None):
    return SwaggerArgumentResponse(
        arg_res.name if new_name is None else new_name,
        arg_res.description if new_description is None else new_description,
        arg_res.arg_type,
        arg_res.example,
        True
    )


def set_nullable(arg_res: Union[SwaggerArgumentResponse, list], condition: Union[str, list] = None,
                 new_name: Union[str, list] = None):
    if isinstance(arg_res, list):
        if new_name is None:
            new_name = [None] * len(arg_res)
        return [set_nullable(cur, condition) for cur, condition, new_name in zip(arg_res, condition, new_name)]
    else:
        return SwaggerArgumentResponse(
            arg_res.name if new_name is None else new_name,
            arg_res.description,
            arg_res.arg_type,
            arg_res.example,
            False,
            condition=condition
        )


def gen_filtered_records(name: str, data: [SwaggerArgumentResponse]):
    wrapper = SwaggerArgumentResponse(
        name=KEY__data,
        description="List of " + name + " records",
        required=True,
        arg_type=SwaggerList(*data)
    )
    return ARG_RES__filtered_records + [wrapper]


ARG_RES__part_sort_pageable = [
    SwaggerArgumentResponse(
        name=KEY__size,
        condition="If not supplied, will return all records",
        description="The size of the page",
        arg_type=int,
        required=False,
        example=[10, 20]
    ),
    SwaggerArgumentResponse(
        name=KEY__page,
        description="The page number to select, 0-based indexing",
        condition="If not supplied will default to 0",
        arg_type=int,
        required=False,
        example=[0, 1]
    )
]


def gen_arg_res_sort_pageable(col_one: str, col_two: str = None, example_one: str = None, example_two: str = None):
    if example_one is None:
        example_one = "jaaql"
    if example_two is None:
        example_two = "jaaql"

    if col_two is None:
        sort_arg = SwaggerArgumentResponse(
            name=KEY__sort,
            description="Comma separated sort",
            condition="If not supplied uses default database ordering",
            arg_type=str,
            required=False,
            example=[col_one + " ASC", col_one + " DESC"]
        )
        search_arg = SwaggerArgumentResponse(
            name=KEY__search,
            description="OR/AND separated search. Uses a limited subset of SQL",
            condition="If not supplied all records will match",
            arg_type=str,
            required=False,
            example=[col_one + " LIKE '%" + example_one + "%'"]
        )
    else:
        sort_arg = SwaggerArgumentResponse(
            name=KEY__sort,
            description="Comma separated sort",
            condition="If not supplied uses default database ordering",
            arg_type=str,
            required=False,
            example=[col_one + " ASC, " + col_two + " DESC", col_two + " ASC, " + col_one + " DESC"]
        )
        search_arg = SwaggerArgumentResponse(
            name=KEY__search,
            description="OR/AND separated search. Uses a limited subset of SQL",
            condition="If not supplied all records will match",
            arg_type=str,
            required=False,
            example=[col_one + " LIKE '%" + example_one + "%' OR " + col_two + " LIKE '%" + example_two + "%'",
                     col_two + " LIKE '%" + example_two + "%' AND " + col_one + " LIKE '%" + example_one + "%'"]
        )

    return [sort_arg, search_arg] + ARG_RES__part_sort_pageable


def combine_response(res: SwaggerResponse, args: Union[SwaggerArgumentResponse, List[SwaggerArgumentResponse]]):
    res = copy.deepcopy(res)
    if not isinstance(args, list):
        args = [args]

    res.responses = res.responses + args
    return res


ARG_RES__username = SwaggerArgumentResponse(
    name=KEY__username,
    description="JAAQL login username",
    arg_type=str,
    lower=True,
    strip=True,
    example=["jaaql", "aaron@jaaql.com"]
)

EXAMPLE__password = ["pa55word", "p@ssword"]

ARG_RES__password = SwaggerArgumentResponse(
    name=KEY__password,
    description="JAAQL login password",
    arg_type=str,
    strip=True,
    example=EXAMPLE__password,
    required=True
)

RES__oauth_token = SwaggerFlatResponse(
    description="A temporary JWT token that can be used to authenticate with the server",
    body=EXAMPLE__jwt
)

DOCUMENTATION__oauth_token = SwaggerDocumentation(
    tags="OAuth",
    security=False,  # This _is_ the security method, therefore it is not expecting a jwt token
    methods=SwaggerMethod(
        name="OAuth Fetch Token",
        description="Authenticate with the server. Send username and password and server will respond with 200 and a "
                    "token which can be used to access the service. The server may also respond with a 202 and a "
                    "token, this indicates that an mfa key is expected. Send the token back to the service along with "
                    "an MFA key and you will returned the aforementioned 200 response",
        method=REST__POST,
        body=[
            ARG_RES__username,
            ARG_RES__password
        ],
        response=RES__oauth_token
    )
)

DOCUMENTATION__oauth_refresh = SwaggerDocumentation(
    tags="OAuth",
    security=True,
    methods=SwaggerMethod(
        name="OAuth Refresh Token",
        description="Refresh your token",
        method=REST__POST,
        response=RES__oauth_token
    )
)



import traceback
from abc import ABC, abstractmethod
from datetime import datetime
import logging
from jaaql.utilities.utils_no_project_imports import time_delta_ms
from jaaql.exceptions.http_status_exception import *
from typing import Optional
import queue

ERR__unknown_echo = "Unknown echo type '%s'. Please use either %s"

KEY_CONFIG__system = "SYSTEM"
KEY_CONFIG__logging = "logging"


RET__echo = "echo"
RET__columns = "columns"
RET__rows = "rows"

DIVIDER__protocol = "://"
DIVIDER__db = "/"
DIVIDER__port = ":"
DIVIDER__password = ":"
DIVIDER__address = "@"

ECHO__none = False
ECHO__debug = "debug"
ECHO__execute = True
ECHO__allowed = [ECHO__none, ECHO__debug, ECHO__execute]

CHAR__newline = "\r\n"

FILE__read = "r"
FILE__query_separator = ";"

ERR__missing_database = "The database name was missing from the connection"


class DBInterface(ABC):

    @abstractmethod
    def __init__(self, config, address: str, username: str):
        self.config = config
        self.output_exceptions = self.config["DEBUG"]["output_query_exceptions"].lower() == "true"
        self.logging = config.get(KEY_CONFIG__system, {KEY_CONFIG__logging: True}).get(KEY_CONFIG__logging, True)
        self.username = username
        self.address = address

    @staticmethod
    def fracture_uri(uri: str, allow_missing_database: bool = False) -> (str, int, str, str, str):
        if DIVIDER__protocol in uri:
            uri = uri.split(DIVIDER__protocol)[1]

        db_split = uri.split(DIVIDER__address)[-1].split(DIVIDER__db)
        address = db_split[0]
        db_name = db_split[1] if len(db_split) > 1 else None

        if not allow_missing_database and db_name is None:
            raise HttpStatusException(ERR__missing_database)

        username, password = DIVIDER__address.join(uri.split(DIVIDER__address)[:-1]).split(DIVIDER__password)
        address, port = address.split(DIVIDER__port)

        return address, port, db_name, username, password

    @abstractmethod
    def get_conn(self):
        pass

    def log_warning(self, exc):
        if self.logging:
            logging.warning(exc, exc_info=False)

    def log_critical(self, exc):
        if self.logging:
            logging.warning(exc, exc_info=False)

    def __attempt_commit_rollback(self, conn, err):
        try:
            if err is None:
                self.commit(conn)
            else:
                self.rollback(conn)
        except Exception as ex:
            if err is None:
                self.log_warning(ex)  # error committing. User stuffed up the SQL
            else:
                self.log_critical(ex)  # error rolling back. It is serious

    def __err_to_exception(self, err, echo):
        if err is not None:
            if isinstance(err, HttpStatusException):
                self.log_warning(err)
                raise err
            else:
                ex = self.handle_db_error(err, echo)
                if ex.response_code in [HTTPStatus.BAD_REQUEST, HTTPStatus.NOT_FOUND, HTTPStatus.UNAUTHORIZED]:
                    self.log_warning(ex)  # Minor error. User made a bad request (bad query, unauthorised etc.)
                else:
                    self.log_warning(ex)  # Serious error, connection failure to db or similar
                raise ex

    def handle_error(self, conn, err, echo=ECHO__none):
        self.__attempt_commit_rollback(conn, err)
        self.__err_to_exception(err, echo)

    def put_conn_handle_error(self, conn, err, echo=ECHO__none, skip_rollback_commit: bool = False):
        if not skip_rollback_commit:
            self.__attempt_commit_rollback(conn, err)

        try:
            self.put_conn(conn)
        except Exception as ex:
            self.log_warning(ex)

        self.__err_to_exception(err, echo)

    @abstractmethod
    def put_conn(self, conn):
        pass

    @abstractmethod
    def rollback(self, conn):
        pass

    def execute_query_fetching_results(self, conn, query, parameters=None, echo=ECHO__none, as_objects=False, wait_hook: queue.Queue = None,
                                       requires_dba_check: bool = False):
        if echo not in ECHO__allowed:
            allowed_echoes = ", ".join([str(allowed_echo) for allowed_echo in ECHO__allowed])
            raise HttpStatusException(ERR__unknown_echo % (str(echo), allowed_echoes), HTTPStatus.BAD_REQUEST)

        if parameters is None:
            parameters = {}

        if echo == ECHO__debug:
            ret = {}
        else:
            new_parameters = {}
            for key, value in parameters.items():
                if value is not None:
                    if isinstance(value, datetime):
                        new_parameters[key] = str(value)
                    else:
                        new_parameters[key] = value

            if requires_dba_check:
                self.check_dba(conn, wait_hook=wait_hook)
                if wait_hook:
                    wait_hook = None

            columns, rows = self.execute_query(conn, query, new_parameters, wait_hook)

            ret = {
                RET__columns: columns,
                RET__rows: rows
            }

        if echo != ECHO__none:
            ret[RET__echo] = query

        if as_objects:
            return self.objectify(ret)
        else:
            return ret

    def objectify(self, data):
        return [dict(zip(data['columns'], row)) for row in data['rows']]

    def execute_script_file(self, conn, file_loc: str = None, as_content: str = None, as_individual=False, commit=True):
        ret = None
        err = None

        if as_individual:
            ret = []

        try:
            if as_content:
                queries = as_content
            else:
                with open(file_loc, FILE__read) as file:
                    queries = file.read()

            if as_individual:
                queries = queries.split(FILE__query_separator)
            else:
                queries = [queries]

            for query in queries:
                if len(query.strip()) != 0:
                    query = query.strip()
                    resp = self.execute_query_fetching_results(conn, query, {}, ECHO__execute)

                    if as_individual:
                        ret.append(resp)
                    else:
                        ret = resp

        except Exception as ex:
            if self.output_exceptions:
                traceback.print_exc()
            err = ex

        if commit:
            self.handle_error(conn, err)
            return ret
        else:
            return ret, err

    @abstractmethod
    def check_dba(self, conn, wait_hook: queue.Queue = None):
        pass

    @abstractmethod
    def execute_query(self, conn, query, parameters: Optional[dict] = None, wait_hook: queue.Queue = None):
        pass

    @abstractmethod
    def commit(self, conn):
        pass

    @abstractmethod
    def handle_db_error(self, err, echo) -> HttpStatusException:
        pass

    @abstractmethod
    def close(self):
        pass
